# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import inspect
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union, get_args

from pydantic import BaseModel, ConfigDict, Field, create_model
from pydantic.fields import FieldInfo

from erniebot_agent.utils.common import create_enum_class
from erniebot_agent.utils.exceptions import RemoteToolError, ToolError

INVALID_FIELD_NAME = "__invalid_field_name__"

_logger = logging.getLogger(__name__)


def is_optional_type(type: Optional[Type]):
    args = get_args(type)
    if len(args) == 0:
        return False

    return len([arg for arg in args if arg is None.__class__]) > 0


def get_typing_list_type(type):
    """get typing.List[T] element type

    Args:
        type (typing.List): Generics type
    """
    # 1. checking list type
    if getattr(type, "_name", None) != "List":
        return None

    arg_type = get_args(type)[0]
    return json_type(arg_type)


def json_type(type: Optional[Type[object]] = None):
    if type is None:
        return "object"

    mapping = {
        int: "integer",
        str: "string",
        list: "array",
        List: "array",
        bytes: "string",
        float: "number",
        ToolParameterView: "object",
    }

    if inspect.isclass(type) and issubclass(type, ToolParameterView):
        return "object"

    if getattr(type, "_name", None) == "List":
        return "array"

    if type not in mapping:
        args = [arg for arg in get_args(type) if arg is not None.__class__]
        if len(args) > 1 or len(args) == 0:
            raise ValueError(
                "only support simple type: FieldType=int/str/float/ToolParameterView, "
                "so the target type should be one of: FieldType, List[FieldType], "
                f"Optional[FieldType], but receive {type}"
            )
        type = args[0]

    if type in mapping:
        return mapping[type]

    if inspect.isclass(type) and issubclass(type, ToolParameterView):
        return "object"

    return str(type)


def python_type_from_json_type(json_type_dict: dict) -> Type[object]:
    simple_types = {"integer": int, "string": str, "number": float, "object": ToolParameterView}
    format = json_type_dict.get("format", None)

    if json_type_dict["type"] == "string" and format == "binary":
        return bytes

    if json_type_dict["type"] in simple_types:
        return simple_types[json_type_dict["type"]]

    assert (
        json_type_dict["type"] == "array"
    ), f"only support simple_types<{','.join(simple_types)}> and array type"
    assert "type" in json_type_dict["items"], "<items> field must be defined when 'type'=array"

    json_type_value = json_type_dict["items"]["type"]
    if json_type_value == "string":
        return List[str]
    if json_type_value == "integer":
        return List[int]
    if json_type_value == "number":
        return List[float]
    if json_type_value == "object":
        return List[ToolParameterView]
    elif json_type_value == "array":
        sub_type = python_type_from_json_type(json_type_dict["items"])
        return List[sub_type]  # type: ignore

    raise ValueError(f"unsupported data type: {json_type_value}")


def scrub_dict(d: dict, remove_empty_dict: bool = False) -> Optional[dict]:
    """remove empty Value node,

        function_call_schema: require

    Args:
        d (dict): the instance of dictionary
        remove_empty_dict (bool): whether remove empty dict

    Returns:
        dict: the dictionary data after slimming down
    """
    if type(d) is dict:
        result = {}
        for k, v in d.items():
            v = scrub_dict(v, remove_empty_dict)
            if v is not None:
                result[k] = v

        if len(result) == 0:
            if not remove_empty_dict:
                return {}
            return None

        return result
    elif isinstance(d, list):
        return [scrub_dict(item, remove_empty_dict) for item in d]  # type: ignore
    else:
        return d


class OpenAPIProperty(BaseModel):
    type: str
    json_schema_extra: Optional[Dict[str, str]] = None
    description: Optional[str] = None
    required: Optional[List[str]] = None
    enum: Optional[List[Union[int, str]]] = None
    items: dict = Field(default_factory=dict)
    properties: dict = Field(default_factory=dict)


def get_field_openapi_property(field_info: FieldInfo) -> OpenAPIProperty:
    """convert pydantic FieldInfo instance to OpenAPIProperty value

    Args:
        field_info (FieldInfo): the field instance

    Returns:
        OpenAPIProperty: the converted OpenAPI Property
    """
    typing_list_type = get_typing_list_type(field_info.annotation)
    if typing_list_type is not None:
        field_type = "array"
    elif field_info.annotation is not None and is_optional_type(field_info.annotation):
        field_type = json_type(get_args(field_info.annotation)[0])
    elif field_info.annotation is not None and issubclass(field_info.annotation, Enum):
        field_type = json_type(type(list(field_info.annotation.__members__.keys())[0]))
    else:
        field_type = json_type(field_info.annotation)

    property: Dict[str, Any] = {
        "type": field_type,
        "description": field_info.description,
    }

    if property["type"] == "array":
        if typing_list_type == "object":
            list_type: Type[ToolParameterView] = get_args(field_info.annotation)[0]
            property["items"] = list_type.to_openapi_dict()
        else:
            if not isinstance(field_info.json_schema_extra, dict):
                raise RemoteToolError("<field_info.json_schema_extra> must be dict data", stage="Loading")

            if "array_items_schema" in field_info.json_schema_extra:
                items_schema: Any = field_info.json_schema_extra["array_items_schema"]

                if not isinstance(items_schema, dict):
                    raise RemoteToolError("<array_items_schema> must be dict data", stage="Loading")

                property["items"] = {
                    "type": items_schema["type"],
                }
                if "description" in items_schema:
                    property["items"]["description"] = items_schema["description"]
            else:
                property["items"] = {"type": typing_list_type}

    elif property["type"] == "object":
        if is_optional_type(field_info.annotation):
            field_type_class: Any = get_args(field_info.annotation)[0]
        else:
            field_type_class = field_info.annotation

        openapi_dict = field_type_class.to_openapi_dict()
        property.update(openapi_dict)
    elif field_info.annotation is not None and issubclass(field_info.annotation, Enum):
        property["enum"] = list(field_info.annotation.__members__.keys())

    property["description"] = property.get("description", "")
    return OpenAPIProperty(**property)


class ToolParameterView(BaseModel):
    __prompt__: Optional[str] = None
    model_config = ConfigDict(use_enum_values=True)

    @classmethod
    def from_openapi_dict(cls, schema: dict) -> Type[ToolParameterView]:
        """parse openapi component schemas to ParameterView
        Args:
            response_or_returns (dict): the content of status code

        Returns:
            _type_: _description_
        """

        # TODO(wj-Mcat): to load Optional field
        fields = {}
        for field_name, field_dict in schema.get("properties", {}).items():
            # skip loading invalid field to improve compatibility
            if "type" not in field_dict:
                raise ToolError(f"`type` field not found in `{field_name}` property", stage="Loading")

            if field_name.startswith("__"):
                continue

            field_type = python_type_from_json_type(field_dict)

            if field_type is List[ToolParameterView]:
                SubParameterView: Type[ToolParameterView] = ToolParameterView.from_openapi_dict(
                    field_dict["items"]
                )
                field_type = List[SubParameterView]  # type: ignore
            elif field_type is ToolParameterView:
                field_type = ToolParameterView.from_openapi_dict(field_dict)
            elif "enum" in field_dict:
                field_type = create_enum_class(field_name, field_dict["enum"])

            # TODO(wj-Mcat): remove supporting for `summary` field
            if "summary" in field_dict:
                description = field_dict["summary"]
                _logger.info("`summary` field will be deprecated, please use `description`")

                if "description" in field_dict:
                    _logger.info("`description` field will be used instead of `summary`")
                    description = field_dict["description"]
            else:
                description = field_dict.get("description", None)

            description = description or ""

            format = field_dict.get("format", None)
            json_schema_extra = {}
            if format is not None:
                json_schema_extra["format"] = format

            json_schema_extra.update(
                {key: value for key, value in field_dict.items() if key.startswith("x-ebagent")}
            )

            if get_typing_list_type(field_type) is not None and field_type is not List[ToolParameterView]:
                json_schema_extra["array_items_schema"] = field_dict["items"]

            field_info_param = dict(
                annotation=field_type, description=description, json_schema_extra=json_schema_extra
            )
            if "default" in field_dict:
                field_info_param["default"] = field_dict["default"]
            field = FieldInfo(**field_info_param)  # type: ignore

            # TODO(wj-Mcat): to handle list field required & not-required
            # if get_typing_list_type(field_type) is not None:
            #     field.default_factory = list

            fields[field_name] = (field_type, field)

        model = create_model("OpenAPIParameterView", __base__=ToolParameterView, **fields)  # type: ignore

        # get the prompt for schema
        model.__prompt__ = schema.get("x-ebagent-prompt", None)
        return model

    @classmethod
    def to_openapi_dict(cls) -> dict:
        """convert ParametersView to openapi spec dict

        Returns:
            dict: schema of openapi
        """

        required_names, properties = [], {}
        for field_name, field_info in cls.model_fields.items():
            if field_info.is_required() and not is_optional_type(field_info.annotation):
                required_names.append(field_name)

            properties[field_name] = dict(get_field_openapi_property(field_info))

        result = {
            "type": "object",
            "properties": properties,
        }
        if len(required_names) > 0:
            result["required"] = required_names
        result = scrub_dict(result, remove_empty_dict=True)  # type: ignore
        return result or {}

    @classmethod
    def function_call_schema(cls) -> dict:
        """get function_call schame

        Returns:
            dict: the schema of function_call
        """
        return cls.to_openapi_dict()

    @classmethod
    def from_dict(cls, field_map: Dict[str, Any]):
        """
        Class method to create a Pydantic model dynamically based on a dictionary.

        Args:
            field_map (Dict[str, Any]): A dictionary mapping field names to their corresponding type
            and description.

        Returns:
            PydanticModel: A dynamically created Pydantic model with fields specified by the
            input dictionary.

        Note:
            This method is used to create a Pydantic model dynamically based on the provided dictionary,
            where each field's type and description are specified in the input.

        """
        fields = {}
        for field_name, field_dict in field_map.items():
            field_type = field_dict["type"]
            description = field_dict["description"]
            field = FieldInfo(annotation=field_type, description=description)
            fields[field_name] = (field_type, field)
        return create_model(cls.__name__, __base__=ToolParameterView, **fields)  # type: ignore


@dataclass
class RemoteToolView:
    uri: str
    method: str
    name: str
    description: str
    version: str

    parameters: Optional[Type[ToolParameterView]] = None
    parameters_description: Optional[str] = None
    parameters_content_type: Optional[str] = None

    returns: Optional[Type[ToolParameterView]] = None
    returns_description: Optional[str] = None
    returns_content_type: Optional[str] = None

    returns_ref_uri: Optional[str] = None
    parameters_ref_uri: Optional[str] = None

    def to_openapi_dict(self):
        result = {
            "operationId": self.name,
            "description": self.description,
        }
        if self.returns is not None:
            response = {
                "200": {
                    "description": self.returns_description,
                    "content": {
                        self.returns_content_type: {
                            "schema": {"$ref": "#/components/schemas/" + (self.returns_ref_uri or "")}
                        }
                    },
                }
            }
            result["responses"] = response

        if self.parameters is not None:
            parameters = {
                "required": True,
                "content": {
                    self.parameters_content_type: {
                        "schema": {"$ref": "#/components/schemas/" + (self.parameters_ref_uri or "")}
                    }
                },
            }
            result["requestBody"] = parameters
        return {self.method: result}

    @staticmethod
    def from_openapi_dict(
        uri: str,
        method: str,
        path_info: dict,
        parameters_views: dict[str, Type[ToolParameterView]],
        version: str,
    ) -> RemoteToolView:
        """construct RemoteToolView from openapi spec-dict info

        Args:
            uri (str): the url path of remote tool
            method (str): http method: one of [get, post, put, delete]
            path_info (dict): the spec info of remote tool
            parameters_views (dict[str, ParametersView]):
                the dict of parameters views which are the schema of input/output of tool
            version (Optional[str]): the optional version of remote tool

        Returns:
            RemoteToolView: the instance of remote tool view
        """
        parameters_ref_uri, returns_ref_uri = None, None
        parameters, parameters_description = None, None
        parameters_content_type, returns_content_type = "application/json", None
        if "requestBody" in path_info:
            request_content = path_info["requestBody"]["content"]
            assert len(request_content.keys()) == 1
            parameters_content_type = list(request_content.keys())[0]
            request_ref = request_content[parameters_content_type]["schema"]["$ref"]
            parameters_ref_uri = request_ref.split("/")[-1]
            assert parameters_ref_uri in parameters_views
            parameters = parameters_views[parameters_ref_uri]
            parameters_description = path_info["requestBody"].get("description", None)

        returns, returns_description = None, None
        if "responses" in path_info:
            response_content = list(path_info["responses"].values())[0]["content"]
            assert len(response_content.keys()) == 1
            returns_content_type = list(response_content.keys())[0]

            # support ref in components.schemas
            if "$ref" in response_content[returns_content_type]["schema"]:
                response_ref = response_content[returns_content_type]["schema"]["$ref"]
                returns_ref_uri = response_ref.split("/")[-1]
                assert returns_ref_uri in parameters_views
                returns = parameters_views[returns_ref_uri]
                returns_description = list(path_info["responses"].values())[0].get("description", None)

        return RemoteToolView(
            name=path_info.get("operationId", uri.strip("/").replace("/", "-")),
            parameters=parameters,
            version=version,
            parameters_description=parameters_description,
            parameters_content_type=parameters_content_type,
            returns=returns,
            returns_description=returns_description,
            returns_content_type=returns_content_type,
            description=path_info.get("description", path_info.get("summary", None)),
            method=method,
            uri=uri,
            # save ref id info
            returns_ref_uri=returns_ref_uri,
            parameters_ref_uri=parameters_ref_uri,
        )

    def function_call_schema(self):
        inputs = {
            "name": self.name,
            "description": self.description,
        }
        if self.parameters is not None:
            inputs["parameters"] = self.parameters.function_call_schema()  # type: ignore
        else:
            inputs["parameters"] = {"type": "object", "properties": {}}

        if self.returns is not None:
            inputs["responses"] = self.returns.function_call_schema()  # type: ignore
        return scrub_dict(inputs) or {}


@dataclass
class Endpoint:
    url: str
    description: Optional[str] = None


@dataclass
class EndpointInfo:
    title: str
    version: str

    description: Optional[str] = None



from typing import Any, Dict, List, Optional

from pydantic import Field

from erniebot_agent.tools.schema import ToolParameterView

from .base import Tool


class LlamaIndexRetrievalToolInputView(ToolParameterView):
    query: str = Field(description="查询语句")
    top_k: int = Field(description="返回结果数量")


class SearchResponseDocument(ToolParameterView):
    title: str = Field(description="检索结果的标题")
    document: str = Field(description="检索结果的内容")


class LlamaIndexRetrievalToolOutputView(ToolParameterView):
    documents: List[SearchResponseDocument] = Field(description="检索结果，内容和用户输入query相关的段落")


class LlamaIndexRetrievalTool(Tool):
    description: str = "在知识库中检索与用户输入query相关的段落"

    def __init__(
        self,
        db,
        threshold: float = 0.0,
        input_type=None,
        output_type=None,
        return_meta_data: bool = True,
    ) -> None:
        super().__init__()
        self.db = db
        self.return_meta_data = return_meta_data
        if input_type is not None:
            self.input_type = input_type
        if output_type is not None:
            self.ouptut_type = output_type
        self.threshold = threshold

    async def __call__(self, query: str, top_k: int = 3, filters: Optional[Dict[str, Any]] = None):
        retriever = self.db.as_retriever(similarity_top_k=top_k)
        nodes = retriever.retrieve(query)
        docs = []
        for doc in nodes:
            if doc.score > self.threshold:
                new_doc = {"content": doc.node.text, "score": doc.score}
                if self.return_meta_data:
                    new_doc["meta"] = doc.metadata
                docs.append(new_doc)
        return {"documents": docs}



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from typing import Dict, List, Optional, Union

from erniebot_agent.memory.messages import FunctionMessage, Message
from erniebot_agent.utils import config_from_environ as C
from erniebot_agent.utils.json import to_pretty_json
from erniebot_agent.utils.output_style import ColoredContent

__all__ = ["set_role_color", "setup_logging"]


def _handle_color_pattern(s: str):
    """Set ASCII color code into right sequence to avoid color conflict."""
    color_pattern = r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])"
    color_lis = re.findall(color_pattern, s)
    origin_text = re.split(color_pattern, s)

    # Preprocess: Split the text by ASCII color code
    idx_color, idx_text = 0, 0
    while idx_text < len(origin_text):
        if idx_text > 0 and origin_text[idx_text - 1] != "" and origin_text[idx_text] != "":
            origin_text.insert(idx_text, "")
        idx_text += 1

    for i in range(len(origin_text)):
        if origin_text[i] == "":
            origin_text[i] = color_lis[idx_color]
            idx_color += 1

    # Process the wrong sequence
    # Set the color after reset code to previous color
    stack: List[str] = []
    for i in range(len(origin_text)):
        if origin_text[i] in color_lis:
            color = origin_text[i]
            if color == "\033[0m":
                stack.pop()
                if stack:
                    origin_text[i] = stack[-1]
            else:
                stack.append(color)
    return "".join(origin_text)


class ColorFormatter(logging.Formatter):
    COLORS = {
        "DEBUG": "\033[95m",  # Purple
        "INFO": "\033[92m",  # Green
        "WARNING": "\033[93m",  # Yellow
        "ERROR": "\033[91m",  # Red
        "CRITICAL": "\033[91m" + "\033[1m",  # Bold Red
        "RESET": "\033[0m",
    }

    def format(self, record):
        # color the message
        origin_args = record.args
        if record.args:
            record_lis = list(record.args)
            for i in range(len(record_lis)):
                if isinstance(record_lis[i], ColoredContent):
                    record_lis[i] = record_lis[i].get_colored_text()
            record.args = tuple(record_lis)

        log_message = super(ColorFormatter, self).format(record)

        log_message = _handle_color_pattern(
            self.COLORS.get(record.levelname, "") + log_message + self.COLORS["RESET"]
        )
        record.args = origin_args
        return log_message


class FileFormatter(logging.Formatter):
    def format(self, record):
        output = []
        if record.args:
            for arg in record.args:
                if isinstance(arg, ColoredContent):
                    output = self.extract_content(arg.text, output)
        log_message = ""
        if output:
            log_message += to_pretty_json(output)
        return log_message

    def extract_content(self, text: Union[Message, str], output: list) -> List[dict]:
        """Extract the content from message and convert to json format."""
        if isinstance(text, Message):
            # Message type
            chat_res, func_res = self.handle_message(text)
            output += [chat_res]
            if func_res:
                output += [{"function_call": [func_res]}]
            return output
        elif isinstance(text, str):
            # Only handle Message Type
            return []

    def handle_message(self, message):
        if isinstance(message, FunctionMessage):
            func_dict = {
                "name": message.name,
                "arguments": message.content,
            }
            return message.to_dict(), func_dict
        else:
            return message.to_dict(), None


def set_role_color(open: bool = True, role_color: Optional[Dict] = None):
    """
    Open or close color role in log, if open, different role will have different color.

    Args:
        open (bool, optional): whether or not to open. Defaults to True.
    """
    if open:
        if not role_color:
            role_color = {"user": "Blue", "function": "Purple", "assistant": "Yellow"}
    else:
        role_color = {}

    ColoredContent.set_global_role_color(role_color)


def setup_logging(
    verbosity: Optional[str] = None,
    use_standard_format: bool = True,
    use_file_handler: bool = False,
    max_log_length: int = 100,
):
    """Configures logging for the ERNIE Bot Agent library.

    Args:
        verbosity: A value indicating the logging level.
        use_standard_format: If True, use the library's standard logging format.
        use_file_handler: If True, use the library's file handler.
        max_log_length: The max length of log message each round.

    Raises:
        ValueError: If the provided verbosity is not a valid logging level.
    """
    logger = logging.getLogger("erniebot_agent")

    if verbosity is None:
        verbosity = C.get_logging_level()

    if verbosity is not None:
        numeric_level = getattr(logging, verbosity.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError(f"Invalid logging level: {verbosity}")

        logger.setLevel(numeric_level)
        logger.propagate = False

        if use_standard_format and not logger.hasHandlers():
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(ColorFormatter("%(levelname)s - %(message)s"))
            logger.addHandler(console_handler)
            set_role_color()

        log_file_path = C.get_logging_file_path()
        if use_file_handler or log_file_path:
            if log_file_path is None:
                log_file_path = "erniebot-agent.log"
            file_handler = logging.FileHandler(log_file_path)
            file_handler.setFormatter(FileFormatter("%(message)s"))
            logger.addHandler(file_handler)

        ColoredContent.set_global_max_length(max_log_length)



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import (
    Any,
    AsyncIterator,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import erniebot
from erniebot import ChatCompletionResponse

from erniebot_agent.chat_models.base import ChatModel
from erniebot_agent.memory.messages import (
    AIMessage,
    AIMessageChunk,
    FunctionCall,
    Message,
    PluginInfo,
    SearchInfo,
    SystemMessage,
)
from erniebot_agent.utils import config_from_environ as C

_T = TypeVar("_T", AIMessage, AIMessageChunk)


class BaseERNIEBot(ChatModel):
    @overload
    async def chat(
        self,
        messages: List[Message],
        *,
        stream: Literal[False] = ...,
        functions: Optional[List[dict]] = ...,
        **kwargs: Any,
    ) -> AIMessage:
        ...

    @overload
    async def chat(
        self,
        messages: List[Message],
        *,
        stream: Literal[True],
        functions: Optional[List[dict]] = ...,
        **kwargs: Any,
    ) -> AsyncIterator[AIMessageChunk]:
        ...

    @overload
    async def chat(
        self, messages: List[Message], *, stream: bool, functions: Optional[List[dict]] = ..., **kwargs: Any
    ) -> Union[AIMessage, AsyncIterator[AIMessageChunk]]:
        ...

    async def chat(
        self,
        messages: List[Message],
        *,
        stream: bool = False,
        functions: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> Union[AIMessage, AsyncIterator[AIMessageChunk]]:
        raise NotImplementedError


class ERNIEBot(BaseERNIEBot):
    """The implementation of the ERNIE Bot model.

    Attributes:
        model (str): The model name.
        api_type (str): The backend of the ERNIE Bot model.
        access_token (Optional[str]): The access token corresponding to the backend.
        default_chat_kwargs (Any): A dict for setting default args for chat model,
            the supported keys include `model`, `_config_`, `top_p`, etc.
    """

    def __init__(
        self,
        model: str,
        api_type: str = "aistudio",
        access_token: Optional[str] = None,
        enable_multi_step_tool_call: bool = False,
        enable_human_clarify: bool = False,
        **default_chat_kwargs: Any,
    ) -> None:
        """Initializes an instance of the `ERNIEBot` class.

        Args:
            model (str): The model name. It should be "ernie-3.5", "ernie-turbo", "ernie-4.0", or
                "ernie-longtext".
            api_type (str): The backend of erniebot. It should be "aistudio" or "qianfan".
                Defaults to "aistudio".
            access_token (Optional[str]): The access token for the backend of erniebot.
                If access_token is None, the global access_token will be used.
            enable_multi_step_tool_call (bool): Whether to enable the multi-step tool call.
                Defaults to False.
            enable_human_clarify (bool): Whether to enable the human clarify. Defaults to False.
            **default_chat_kwargs: Keyword arguments, such as `_config_`, `top_p`, `temperature`,
                `penalty_score`, and `system`.
        """
        super().__init__(model=model, **default_chat_kwargs)

        self.api_type = api_type
        if access_token is None:
            access_token = C.get_global_access_token()
        self.access_token = access_token
        self._maybe_validate_qianfan_auth()

        self.extra_data = {}
        self.extra_data["multi_step_tool_call_close"] = not enable_multi_step_tool_call
        self.extra_data["chat_with_human_close"] = not enable_human_clarify

    @overload
    async def chat(
        self,
        messages: List[Message],
        *,
        stream: Literal[False] = ...,
        functions: Optional[List[dict]] = ...,
        **kwargs: Any,
    ) -> AIMessage:
        ...

    @overload
    async def chat(
        self,
        messages: List[Message],
        *,
        stream: Literal[True],
        functions: Optional[List[dict]] = ...,
        **kwargs: Any,
    ) -> AsyncIterator[AIMessageChunk]:
        ...

    @overload
    async def chat(
        self, messages: List[Message], *, stream: bool, functions: Optional[List[dict]] = ..., **kwargs: Any
    ) -> Union[AIMessage, AsyncIterator[AIMessageChunk]]:
        ...

    async def chat(
        self,
        messages: List[Message],
        *,
        stream: bool = False,
        functions: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> Union[AIMessage, AsyncIterator[AIMessageChunk]]:
        """Asynchronously chats with the ERNIE Bot model.

        Args:
            messages (List[Message]): A list of messages.
            stream (bool): Whether to use streaming generation. Defaults to False.
            functions (Optional[List[dict]]): The function definitions to be used by the model.
                Defaults to None.
            **kwargs: Keyword arguments, such as `top_p`, `temperature`, `penalty_score`, and `system`.

        Returns:
            If `stream` is False, returns a single message.
            If `stream` is True, returns an asynchronous iterator of message chunks.
        """

        cfg_dict = self._generate_config(messages, functions=functions, **kwargs)

        response = await self._generate_response(cfg_dict, stream, functions)

        if not stream:
            assert isinstance(response, ChatCompletionResponse)
            return convert_response_to_output(response, AIMessage)
        else:
            assert isinstance(response, AsyncIterator)
            # We have to do type casting here due to the known mypy issue:
            # https://github.com/python/mypy/issues/16590
            return (
                convert_response_to_output(resp, AIMessageChunk)
                async for resp in cast(AsyncIterator[ChatCompletionResponse], response)
            )

    def _generate_config(self, messages: List[Message], functions, **kwargs) -> dict:
        if any(isinstance(m, SystemMessage) for m in messages):
            raise ValueError(f"The input messages should not contain SystemMessage: {messages}")
        cfg_dict = self.default_chat_kwargs.copy()

        cfg_dict.setdefault("_config_", {})
        if self.api_type is not None:
            cfg_dict["_config_"]["api_type"] = self.api_type
        if hasattr(self, "ak") and hasattr(self, "sk"):
            cfg_dict["_config_"]["ak"] = self.ak
            cfg_dict["_config_"]["sk"] = self.sk
        cfg_dict["_config_"]["access_token"] = self.access_token

        cfg_dict["messages"] = [m.to_dict() for m in messages]
        cfg_dict["model"] = self.model
        if functions is not None:
            cfg_dict["functions"] = functions

        name_list = ["top_p", "temperature", "penalty_score", "system", "plugins", "tool_choice"]
        for name in name_list:
            if name in kwargs:
                cfg_dict[name] = kwargs[name]

        if "plugins" in cfg_dict and (cfg_dict["plugins"] is None or len(cfg_dict["plugins"]) == 0):
            cfg_dict.pop("plugins")

        if "tool_choice" in cfg_dict:
            # rm blank dict
            if not cfg_dict["tool_choice"]:
                cfg_dict.pop("tool_choice")
        return cfg_dict

    def _maybe_validate_qianfan_auth(self) -> None:
        if self.api_type == "qianfan":
            if self.access_token is None:
                # 默认选择千帆时，如果设置了access_token，这个access_token不是aistudio的
                if "ak" and "sk" not in self.default_chat_kwargs:
                    ak, sk = C.get_global_aksk()
                    if ak is None or sk is None:
                        raise RuntimeError("Please set at least one of ak+sk or access token.")
                    else:
                        self.ak = ak
                        self.sk = sk
                else:
                    self.ak = self.default_chat_kwargs.pop("ak")
                    self.sk = self.default_chat_kwargs.pop("sk")
            else:
                # If set access_token in environment and pass ak and sk in default_chat_kwargs,
                # the access_token in default_chat_kwargs will be used.
                if "ak" and "sk" in self.default_chat_kwargs:
                    self.ak = self.default_chat_kwargs.pop("ak")
                    self.sk = self.default_chat_kwargs.pop("sk")

    async def _generate_response(
        self, cfg_dict: dict, stream: bool, functions: Optional[List[dict]]
    ) -> Union[ChatCompletionResponse, AsyncIterator[ChatCompletionResponse]]:
        # TODO: Improve this when erniebot typing issue is fixed.
        # Note: If plugins is not None, erniebot will not use Baidu_search.
        if "plugins" in cfg_dict:
            response = await erniebot.ChatCompletionWithPlugins.acreate(
                messages=cfg_dict["messages"],
                plugins=cfg_dict["plugins"],  # type: ignore
                stream=stream,
                _config_=cfg_dict["_config_"],
                functions=functions,  # type: ignore
                extra_params={
                    "extra_data": json.dumps(self.extra_data),
                },
            )
        else:
            response = await erniebot.ChatCompletion.acreate(
                stream=stream,
                extra_params={
                    "extra_data": json.dumps(self.extra_data),
                },
                **cfg_dict,
            )

        return response


def convert_response_to_output(response: ChatCompletionResponse, output_type: Type[_T]) -> _T:
    clarify = False
    # ernie-turbo has no `finish_reason`
    if hasattr(response, "finish_reason") and response["finish_reason"] == "plugin_clarify":
        clarify = True

    if hasattr(response, "function_call"):
        function_call = FunctionCall(
            name=response.function_call["name"],
            thoughts=response.function_call["thoughts"],
            arguments=response.function_call["arguments"],
        )
        return output_type(
            content="",
            function_call=function_call,
            plugin_info=None,
            search_info=None,
            clarify=clarify,
            token_usage=response.usage,
        )
    elif hasattr(response, "plugin_info"):
        plugin_info = PluginInfo(
            names=[
                response["plugin_metas"][i]["pluginNameForModel"]
                for i in range(len(response["plugin_metas"]))
            ]
        )

        return output_type(
            content=response.result,
            function_call=None,
            plugin_info=plugin_info,
            search_info=None,
            token_usage=response.usage,
            clarify=clarify,
        )
    elif hasattr(response, "search_info") and len(response.search_info.items()) > 0:
        search_info = SearchInfo(
            results=response.search_info["search_results"],
        )
        return output_type(
            content=response.result,
            function_call=None,
            plugin_info=None,
            search_info=search_info,
            token_usage=response.usage,
            clarify=clarify,
        )
    else:
        return output_type(
            content=response.result,
            function_call=None,
            plugin_info=None,
            search_info=None,
            token_usage=response.usage,
            clarify=clarify,
        )



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (
    Any,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    runtime_checkable,
)

from erniebot_agent.agents.schema import AgentResponse, LLMResponse, ToolResponse
from erniebot_agent.chat_models.base import ChatModel
from erniebot_agent.file import File, FileManager
from erniebot_agent.memory import Memory
from erniebot_agent.memory.messages import Message
from erniebot_agent.tools.base import BaseTool

LLMT = TypeVar("LLMT", bound=ChatModel)


class BaseAgent(Protocol[LLMT]):
    llm: LLMT
    memory: Memory

    async def run(self, prompt: str, files: Optional[Sequence[File]] = None) -> AgentResponse:
        ...

    async def run_llm(
        self,
        messages: List[Message],
        *,
        llm_opts: Optional[Mapping[str, Any]] = None,
    ) -> LLMResponse:
        ...

    async def run_tool(self, tool_name: str, tool_args: str) -> ToolResponse:
        ...

    def load_tool(self, tool: BaseTool) -> None:
        ...

    def unload_tool(self, tool: BaseTool) -> None:
        ...

    def get_tool(self, tool_name: str) -> BaseTool:
        ...

    def get_tools(self) -> List[BaseTool]:
        ...

    def reset_memory(self) -> None:
        ...

    def get_file_manager(self) -> FileManager:
        ...


@runtime_checkable
class AgentLike(Protocol):
    async def run(self, prompt: str, files: Optional[Sequence[File]] = None) -> AgentResponse:
        ...



from typing import Dict, Type

from pydantic import Field

from erniebot_agent.chat_models import ERNIEBot
from erniebot_agent.memory.messages import HumanMessage
from erniebot_agent.tools.base import Tool
from erniebot_agent.tools.schema import ToolParameterView


class ChatWithEBInputView(ToolParameterView):
    query: str = Field(description="需要询问EB的具体问题")


class ChatWithEBOutputView(ToolParameterView):
    response: str = Field(description="EB回复的结果")


class ChatWithEB(Tool):
    description: str = (
        "ChatWithEB是一款根据用户的问题，向EB生成式大语言模型进行提问，并获取EB回答结果的工具。EB一般能解决知识型问答、文本创作、信息查询、信息检索等基础的文本生成和信息检索功能"
    )
    input_type: Type[ToolParameterView] = ChatWithEBInputView
    ouptut_type: Type[ToolParameterView] = ChatWithEBOutputView

    def __init__(self, llm: ERNIEBot):
        self.llm = llm

    async def __call__(self, query: str) -> Dict[str, str]:
        response = await self.llm.chat([HumanMessage(query)])
        return {"response": response.content}



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import warnings
from typing import Any, NoReturn, Protocol, final

from erniebot_agent.utils.exceptions import ObjectClosedError


class Closeable(Protocol):
    @property
    def closed(self) -> bool:
        ...

    def __del__(self, _warn=warnings.warn) -> None:
        if not self.closed:
            _warn(f"Unclosed object: {repr(self)}", ResourceWarning, source=self)

    async def close(self) -> None:
        ...

    async def aclose(self) -> None:
        return await self.close()

    def ensure_not_closed(self) -> None:
        if self.closed:
            raise ObjectClosedError(f"{repr(self)} is closed.")


class Noncopyable(object):
    @final
    def __copy__(self) -> NoReturn:
        raise TypeError("Cannot copy an instance of a non-copyable class.")

    @final
    def __deepcopy__(self, memo: Any) -> NoReturn:
        raise TypeError("Cannot deep-copy an instance of a non-copyable class.")

    @final
    def __reduce__(self) -> NoReturn:
        raise TypeError("Cannot pickle an instance of a non-copyable class.")

    @final
    def __reduce_ex__(self, protocol: Any) -> NoReturn:
        raise TypeError("Cannot pickle an instance of a non-copyable class.")



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import functools
import json
import types
from typing import Callable, Dict, Iterable, List, final

from erniebot_agent.tools.base import BaseTool, Tool
from erniebot_agent.tools.utils import get_fastapi_openapi


@final
class ToolManager(object):
    """A `ToolManager` instance manages tools for an agent.

    This implementation is based on `ToolsManager` in
    https://github.com/deepset-ai/haystack/blob/main/haystack/agents/base.py
    """

    def __init__(self, tools: Iterable[BaseTool]) -> None:
        super().__init__()
        self._tools: Dict[str, BaseTool] = {}
        for tool in tools:
            self.add_tool(tool)

    def __getitem__(self, tool_name: str) -> BaseTool:
        return self.get_tool(tool_name)

    def add_tool(self, tool: BaseTool) -> None:
        tool_name = tool.tool_name
        if tool_name in self._tools:
            raise ValueError(f"Name {repr(tool_name)} is already registered.")
        self._tools[tool_name] = tool

    def remove_tool(self, tool: BaseTool) -> None:
        tool_name = tool.tool_name
        if tool_name not in self._tools:
            raise ValueError(f"Name {repr(tool_name)} is not registered.")
        if self._tools[tool_name] is not tool:
            raise RuntimeError(f"The tool with the registered name {repr(tool_name)} is not the given tool.")
        self._tools.pop(tool_name)

    def get_tool(self, tool_name: str) -> BaseTool:
        if tool_name not in self._tools:
            raise ValueError(f"Name {repr(tool_name)} is not registered.")
        return self._tools[tool_name]

    def get_tools(self) -> List[BaseTool]:
        return list(self._tools.values())

    def get_tool_names(self) -> str:
        return ", ".join(self._tools.keys())

    def get_tool_names_with_descriptions(self) -> str:
        return "\n".join(
            f"{name}:{json.dumps(tool.function_call_schema())}" for name, tool in self._tools.items()
        )

    def get_tool_schemas(self):
        return [tool.function_call_schema() for tool in self._tools.values()]

    def serve(self, port: int = 5000):
        """start the local server for toolkit

        Args:
            port (int, optional): the port of local toolkit server. Defaults to 5000.
        """
        try:
            import uvicorn
            from fastapi import FastAPI
        except ImportError:
            raise ImportError(
                "Could not import fastapi or uvicorn python package. Please install it "
                "with `pip install uvicorn fastapi`."
            )

        app = FastAPI(title="erniebot-agent-tools", version="0.0")

        def create_func(f: Callable, func_types: dict, tool: BaseTool) -> Callable:
            """create new function for fastapi routers

            Args:
                f (Callable): the proxy function to call tool
                func_types (dict): the annotations for function inputs
                tool (BaseTool): the instance of the tool
            """
            new_func = types.FunctionType(
                f.__code__, f.__globals__, f.__name__, f.__defaults__, f.__closure__
            )
            new_func.__annotations__ = func_types
            return functools.partial(new_func, __tool__=tool)

        for tool in self._tools.values():
            if not isinstance(tool, Tool):
                continue

            async def create_tool_endpoint_without_inputs(__tool__):
                return await __tool__()

            async def create_tool_endpoint(__tool__, inputs):
                data = inputs.model_dump(mode="json")
                return await __tool__(**data)

            if tool.input_type is not None:
                type_annotation = {"inputs": tool.input_type}
                func = create_func(create_tool_endpoint, type_annotation, tool)
            else:
                func = create_func(create_tool_endpoint_without_inputs, {}, tool)

            tool_name = tool.tool_name.split("/")[-1]
            app.add_api_route(
                f"/erniebot-agent-tools/0.0/{tool_name}",
                endpoint=func,
                response_model=tool.ouptut_type,
                description=tool.description,
                operation_id=tool.tool_name,
                methods=["POST"],
            )

        @app.get("/.well-known/openapi.yaml")
        def get_openapi_yaml():
            """get openapi json schema from fastapi"""
            return app.openapi()

        get_fastapi_openapi(app)
        uvicorn.run(app, host="0.0.0.0", port=port)



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import uuid
from typing import Any, Dict, List, Optional, Type

import erniebot
from pydantic import Field

from erniebot_agent.memory.messages import AIMessage, HumanMessage, Message
from erniebot_agent.tools.base import Tool
from erniebot_agent.tools.schema import ToolParameterView
from erniebot_agent.utils.common import download_file, get_cache_dir


class ImageGenerationInputView(ToolParameterView):
    prompt: str = Field(description="描述图像内容、风格的文本。例如：生成一张月亮的照片，月亮很圆。")
    width: int = Field(description="生成图片的宽度")
    height: int = Field(description="生成图片的高度")
    image_num: int = Field(description="生成图片的数量")


class ImageGenerationOutputView(ToolParameterView):
    image_path: str = Field(description="图片在本地机器上的保存路径")


class ImageGenerationTool(Tool):
    description: str = "AI作图、生成图片、画图的工具"
    input_type: Type[ToolParameterView] = ImageGenerationInputView
    ouptut_type: Type[ToolParameterView] = ImageGenerationOutputView

    def __init__(
        self,
        yinian_access_token: Optional[str] = None,
        yinian_ak: Optional[str] = None,
        yinian_sk: Optional[str] = None,
    ) -> None:
        self.config: Dict[str, Optional[Any]]
        if yinian_access_token is not None:
            self.config = {"api_type": "yinian", "access_token": yinian_access_token}
        elif yinian_ak is not None and yinian_sk is not None:
            self.config = {"api_type": "yinian", "ak": yinian_ak, "sk": yinian_sk}
        else:
            raise ValueError("Please set the yinian_access_token, or set yinian_ak and yinian_sk")

    async def __call__(
        self,
        prompt: str,
        width: int = 512,
        height: int = 512,
        image_num: int = 1,
    ) -> Dict[str, List[str]]:
        response = erniebot.Image.create(  # type: ignore
            model="ernie-vilg-v2",
            prompt=prompt,
            width=width,
            height=height,
            image_num=image_num,
            _config_=self.config,
        )

        image_path = []
        cache_dir = get_cache_dir()
        for item in response["data"]["sub_task_result_list"]:
            image_url = item["final_image_list"][0]["img_url"]
            save_path = os.path.join(cache_dir, f"img_{uuid.uuid1()}.png")
            download_file(image_url, save_path)
            image_path.append(save_path)
        return {"image_path": image_path}

    @property
    def examples(self) -> List[Message]:
        return [
            HumanMessage("画一张小狗的图片，图像高度512，图像宽度512"),
            AIMessage(
                "",
                function_call={
                    "name": "ImageGenerationTool",
                    "thoughts": "用户需要我生成一张小狗的图片，图像高度为512，宽度为512。我可以使用ImageGenerationTool工具来满足用户的需求。",
                    "arguments": '{"prompt":"画一张小狗的图片，图像高度512，图像宽度512",'
                    '"width":512,"height":512,"image_num":1}',
                },
            ),
            HumanMessage("生成两张天空的图片"),
            AIMessage(
                "",
                function_call={
                    "name": self.tool_name,
                    "thoughts": "用户想要生成两张天空的图片，我需要调用ImageGenerationTool工具的call接口，"
                    "并设置prompt为'生成两张天空的图片'，width和height可以默认为512，image_num默认为2。",
                    "arguments": '{"prompt":"生成两张天空的图片","width":512,"height":512,"image_num":2}',
                },
            ),
            HumanMessage("使用AI作图工具，生成1张小猫的图片，高度和高度是1024"),
            AIMessage(
                "",
                function_call={
                    "name": self.tool_name,
                    "thoughts": "用户需要生成一张小猫的图片，高度和宽度都是1024。我可以使用ImageGenerationTool工具来满足用户的需求。",
                    "arguments": '{"prompt":"生成一张小猫的照片。","width":1024,"height":1024,"image_num":1}',
                },
            ),
        ]



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import requests


def is_url(path):
    """
    Whether path is URL.
    Args:
        path (string): URL string or not.
    """
    return path.startswith("http://") or path.startswith("https://")


def url_file_exists(url: str, headers: dict) -> bool:
    """check whether the url file exists

        refer to: https://stackoverflow.com/questions/2486145/python-check-if-url-to-jpg-exists

    Args:
        url (str): the url of target file
        headers (dict): the header of http request

    Returns:
        bool: whether the url file exists
    """
    if not is_url(url):
        return False

    result = requests.head(url, headers=headers)
    return result.status_code == requests.codes.ok



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Optional

from erniebot_agent.agents.base import BaseAgent
from erniebot_agent.agents.callback.handlers.base import CallbackHandler
from erniebot_agent.agents.schema import AgentResponse, LLMResponse, ToolResponse
from erniebot_agent.chat_models.base import ChatModel
from erniebot_agent.memory.messages import Message
from erniebot_agent.tools.base import BaseTool
from erniebot_agent.utils.json import to_pretty_json
from erniebot_agent.utils.output_style import ColoredContent

_logger = logging.getLogger(__name__)


class LoggingHandler(CallbackHandler):
    """A callback handler for logging."""

    logger: logging.Logger

    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """Initialize a logging handler.

        Args:
            logger: The logger to use. If `None`, a default logger will be used.
        """
        super().__init__()

        if logger is None:
            self.logger = _logger
        else:
            self.logger = logger

    async def on_run_start(self, agent: BaseAgent, prompt: str) -> None:
        """Called to log when the agent starts running."""
        self._agent_info(
            "%s is about to start running with input:\n%s",
            agent.__class__.__name__,
            ColoredContent(prompt, role="user"),
            subject="Run",
            state="Start",
        )

    async def on_llm_start(self, agent: BaseAgent, llm: ChatModel, messages: List[Message]) -> None:
        """Called to log when the LLM starts running."""
        # TODO: Prettier messages
        self._agent_info(
            "%s is about to start running with input:\n%s",
            llm.__class__.__name__,
            ColoredContent(messages[-1]),
            subject="LLM",
            state="Start",
        )

    async def on_llm_end(self, agent: BaseAgent, llm: ChatModel, response: LLMResponse) -> None:
        """Called to log when the LLM successfully ends running."""
        self._agent_info(
            "%s finished running with output:\n%s",
            llm.__class__.__name__,
            ColoredContent(response.message),
            subject="LLM",
            state="End",
        )

    async def on_tool_start(self, agent: BaseAgent, tool: BaseTool, input_args: str) -> None:
        """Called to log when a tool starts running."""
        js_inputs = to_pretty_json(input_args, from_json=True)
        self._agent_info(
            "%s is about to start running with input:\n%s",
            ColoredContent(tool.__class__.__name__, role="function"),
            ColoredContent(js_inputs, role="function"),
            subject="Tool",
            state="Start",
        )

    async def on_tool_end(self, agent: BaseAgent, tool: BaseTool, response: ToolResponse) -> None:
        """Called to log when a tool successfully ends running."""
        js_inputs = to_pretty_json(response.json, from_json=True)
        self._agent_info(
            "%s finished running with output:\n%s",
            ColoredContent(tool.__class__.__name__, role="function"),
            ColoredContent(js_inputs, role="function"),
            subject="Tool",
            state="End",
        )

    async def on_run_end(self, agent: BaseAgent, response: AgentResponse) -> None:
        """Called to log when the agent successfully ends running."""
        self._agent_info("%s finished running.", agent.__class__.__name__, subject="Run", state="End")

    def _agent_info(self, msg: str, *args, subject, state, **kwargs) -> None:
        msg = f"[{subject}][{state}] {msg}"
        self.logger.info(msg, *args, **kwargs)



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Any, AsyncIterator, List, Literal, Union, overload

from erniebot_agent.memory.messages import AIMessage, AIMessageChunk, Message


class ChatModel(metaclass=ABCMeta):
    """The base class of chat-optimized LLM.

    Attributes:
        model (str): The model name.
        default_chat_kwargs (Any): A dict for setting default args for chat model,
            the supported keys include `model`, `_config_`, `top_p`, etc.
    """

    def __init__(self, model: str, **default_chat_kwargs: Any):
        self.model = model
        self.default_chat_kwargs = default_chat_kwargs

    @overload
    async def chat(
        self, messages: List[Message], *, stream: Literal[False] = ..., **kwargs: Any
    ) -> AIMessage:
        ...

    @overload
    async def chat(
        self, messages: List[Message], *, stream: Literal[True], **kwargs: Any
    ) -> AsyncIterator[AIMessageChunk]:
        ...

    @overload
    async def chat(
        self, messages: List[Message], *, stream: bool, **kwargs: Any
    ) -> Union[AIMessage, AsyncIterator[AIMessageChunk]]:
        ...

    @abstractmethod
    async def chat(
        self, messages: List[Message], *, stream: bool = False, **kwargs: Any
    ) -> Union[AIMessage, AsyncIterator[AIMessageChunk]]:
        """The abstract method for asynchronously chatting with the LLM.

        Args:
            messages (List[Message]): A list of messages.
            stream (bool): Whether to use streaming generation. Defaults to False.
            **kwargs: Keyword arguments, such as `top_p`, `temperature`, `penalty_score`, and `system`.

        Returns:
            If stream is False, returns a single message.
            If stream is True, returns an asynchronous iterator of message chunks.
        """
        raise NotImplementedError



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import re
from typing import Final, Generator, List, Literal, get_args

from typing_extensions import TypeAlias

_LOCAL_FILE_ID_PREFIX: Final[str] = "file-local-"
_UUID_PATTERN: Final[str] = r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
_LOCAL_FILE_ID_PATTERN: Final[str] = _LOCAL_FILE_ID_PREFIX + _UUID_PATTERN
_REMOTE_FILE_ID_PREFIX: Final[str] = "file-"
_REMOTE_FILE_ID_PATTERN: Final[str] = _REMOTE_FILE_ID_PREFIX + r"[0-9]{15}"

FilePurpose: TypeAlias = Literal["assistants", "assistants_output"]

_compiled_local_file_id_pattern = re.compile(_LOCAL_FILE_ID_PATTERN)
_compiled_remote_file_id_pattern = re.compile(_REMOTE_FILE_ID_PATTERN)


def create_local_file_id_from_uuid(uuid: str) -> str:
    """Create a random local file id."""
    return _LOCAL_FILE_ID_PREFIX + uuid


def get_timestamp() -> str:
    return datetime.datetime.now().isoformat(sep=" ", timespec="seconds")


def is_file_id(str_: str) -> bool:
    """Judge whether a file id is valid or not."""
    return is_local_file_id(str_) or is_remote_file_id(str_)


def is_local_file_id(str_: str) -> bool:
    """Judge whether a file id is a valid local file id or not."""
    return _compiled_local_file_id_pattern.fullmatch(str_) is not None


def is_remote_file_id(str_: str) -> bool:
    """Judge whether a file id is a valid remote file id or not."""
    return _compiled_remote_file_id_pattern.fullmatch(str_) is not None


def extract_file_ids(str_: str) -> List[str]:
    """Find all file ids in a string."""
    return extract_local_file_ids(str_) + extract_remote_file_ids(str_)


def extract_local_file_ids(str_: str) -> List[str]:
    """Find all local file ids in a string."""
    return _compiled_local_file_id_pattern.findall(str_)


def extract_remote_file_ids(str_: str) -> List[str]:
    """Find all remote file ids in a string."""
    return _compiled_remote_file_id_pattern.findall(str_)


def is_valid_file_purpose(file_purpose: str) -> bool:
    return file_purpose in get_args(FilePurpose)


def generate_fake_remote_file_ids() -> Generator[str, None, None]:
    counter = 0
    while True:
        number = f"{counter:015d}"
        if len(number) > 15:
            break
        yield _REMOTE_FILE_ID_PREFIX + number



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Final, Iterable, List, Optional, Sequence, Tuple, Union

from erniebot_agent.agents.agent import Agent
from erniebot_agent.agents.callback.callback_manager import CallbackManager
from erniebot_agent.agents.callback.handlers.base import CallbackHandler
from erniebot_agent.agents.schema import (
    DEFAULT_FINISH_STEP,
    AgentResponse,
    AgentStep,
    EndInfo,
    EndStep,
    PluginStep,
    ToolInfo,
    ToolStep,
)
from erniebot_agent.chat_models.erniebot import BaseERNIEBot
from erniebot_agent.file import File, FileManager
from erniebot_agent.memory import Memory
from erniebot_agent.memory.messages import FunctionMessage, HumanMessage, Message
from erniebot_agent.tools.base import BaseTool
from erniebot_agent.tools.tool_manager import ToolManager

_MAX_STEPS: Final[int] = 5
_logger = logging.getLogger(__name__)


class FunctionAgent(Agent):
    """An agent driven by function calling.

    The orchestration capabilities of a function agent are powered by the
    function calling ability of LLMs. Typically, a function agent asks the LLM
    to generate a response that can be parsed into an action (e.g., calling a
    tool with given arguments), and then the agent takes that action, which
    forms an agent step. The agent repeats this process until the maximum number
    of steps is reached or the LLM considers the task finished.

    Attributes:
        llm: The LLM that the agent uses.
        memory: The message storage that keeps the chat history.
        max_steps: The maximum number of steps in each agent run.
    """

    llm: BaseERNIEBot
    memory: Memory
    max_steps: int

    def __init__(
        self,
        llm: BaseERNIEBot,
        tools: Union[ToolManager, Iterable[BaseTool]],
        *,
        memory: Optional[Memory] = None,
        system: Optional[str] = None,
        callbacks: Optional[Union[CallbackManager, Iterable[CallbackHandler]]] = None,
        file_manager: Optional[FileManager] = None,
        plugins: Optional[List[str]] = None,
        max_steps: Optional[int] = None,
        first_tools: Optional[Sequence[BaseTool]] = [],
    ) -> None:
        """Initialize a function agent.

        Args:
            llm: An LLM for the agent to use.
            tools: A list of tools for the agent to use.
            memory: A memory object that equips the agent to remember chat
                history. If `None`, a `WholeMemory` object will be used.
            system: A message that tells the LLM how to interpret the
                conversations. If `None`, the system message contained in
                `memory` will be used.
            callbacks: A list of callback handlers for the agent to use. If
                `None`, a default list of callbacks will be used.
            file_manager: A file manager for the agent to interact with files.
                If `None`, a global file manager that can be shared among
                different components will be implicitly created and used.
            plugins: A list of names of the plugins for the agent to use. If
                `None`, the agent will use a default list of plugins. Set
                `plugins` to `[]` to disable the use of plugins.
            max_steps: The maximum number of steps in each agent run. If `None`,
                use a default value.
            first_tools: Tools scheduled to be called sequentially at the
                beginning of each agent run.

        Raises:
            ValueError: if `max_steps` is non-positive.
            RuntimeError: if tools in first_tools but not in tools list.

        """
        super().__init__(
            llm=llm,
            tools=tools,
            memory=memory,
            system=system,
            callbacks=callbacks,
            file_manager=file_manager,
            plugins=plugins,
        )
        if max_steps is not None:
            if max_steps <= 0:
                raise ValueError("Invalid `max_steps` value")
            self.max_steps = max_steps
        else:
            self.max_steps = _MAX_STEPS

        if first_tools:
            self._first_tools = first_tools
            for tool in self._first_tools:
                if tool not in self.get_tools():
                    raise RuntimeError("The tool in `first_tools` must be in the tools list.")
        else:
            self._first_tools = []

    async def _run(self, prompt: str, files: Optional[Sequence[File]] = None) -> AgentResponse:
        chat_history: List[Message] = []
        steps_taken: List[AgentStep] = []

        run_input = await HumanMessage.create_with_files(
            prompt, files or [], include_file_urls=self.file_needs_url
        )

        num_steps_taken = 0
        chat_history.append(run_input)

        for tool in self._first_tools:
            curr_step, new_messages = await self._step(chat_history, selected_tool=tool)
            if not isinstance(curr_step, EndStep):
                chat_history.extend(new_messages)
                num_steps_taken += 1
                steps_taken.append(curr_step)
            else:
                # If tool choice not work, skip this round
                _logger.warning(f"Selected tool [{tool.tool_name}] not work")

        while num_steps_taken < self.max_steps:
            curr_step, new_messages = await self._step(chat_history)
            chat_history.extend(new_messages)
            if isinstance(curr_step, ToolStep):
                steps_taken.append(curr_step)

            elif isinstance(curr_step, PluginStep):
                steps_taken.append(curr_step)
                # 预留 调用了Plugin之后不结束的接口

                # 此处为调用了Plugin之后直接结束的Plugin
                curr_step = DEFAULT_FINISH_STEP

            if isinstance(curr_step, EndStep):
                response = self._create_finished_response(chat_history, steps_taken, curr_step)
                self.memory.add_message(chat_history[0])
                self.memory.add_message(chat_history[-1])
                return response
            num_steps_taken += 1
        response = self._create_stopped_response(chat_history, steps_taken)
        return response

    async def _step(
        self, chat_history: List[Message], selected_tool: Optional[BaseTool] = None
    ) -> Tuple[AgentStep, List[Message]]:
        new_messages: List[Message] = []
        input_messages = self.memory.get_messages() + chat_history
        if selected_tool is not None:
            tool_choice = {"type": "function", "function": {"name": selected_tool.tool_name}}
            llm_resp = await self.run_llm(
                messages=input_messages,
                functions=[selected_tool.function_call_schema()],  # only regist one tool
                tool_choice=tool_choice,
            )
        else:
            llm_resp = await self.run_llm(messages=input_messages)

        output_message = llm_resp.message  # AIMessage
        new_messages.append(output_message)
        if output_message.function_call is not None:
            tool_name = output_message.function_call["name"]
            tool_args = output_message.function_call["arguments"]
            tool_resp = await self.run_tool(tool_name=tool_name, tool_args=tool_args)
            new_messages.append(FunctionMessage(name=tool_name, content=tool_resp.json))
            return (
                ToolStep(
                    info=ToolInfo(tool_name=tool_name, tool_args=tool_args),
                    result=tool_resp.json,
                    input_files=tool_resp.input_files,
                    output_files=tool_resp.output_files,
                ),
                new_messages,
            )
        elif output_message.plugin_info is not None:
            file_manager = self.get_file_manager()
            return (
                PluginStep(
                    info=output_message.plugin_info,
                    result=output_message.content,
                    input_files=file_manager.sniff_and_extract_files_from_text(
                        chat_history[-1].content
                    ),  # TODO: make sure this is correct.
                    output_files=file_manager.sniff_and_extract_files_from_text(output_message.content),
                ),
                new_messages,
            )
        else:
            if output_message.clarify:
                # `clarify` and [`function_call`, `plugin`(directly end)] will not appear at the same time
                return EndStep(info=EndInfo(end_reason="CLARIFY"), result=None), new_messages
            return DEFAULT_FINISH_STEP, new_messages

    def _create_finished_response(
        self,
        chat_history: List[Message],
        steps: List[AgentStep],
        curr_step: EndStep,
    ) -> AgentResponse:
        last_message = chat_history[-1]
        return AgentResponse(
            text=last_message.content,
            chat_history=chat_history,
            steps=steps,
            status=curr_step.info["end_reason"],
        )

    def _create_stopped_response(
        self,
        chat_history: List[Message],
        steps: List[AgentStep],
    ) -> AgentResponse:
        return AgentResponse(
            text="Agent run stopped early.",
            chat_history=chat_history,
            steps=steps,
            status="STOPPED",
        )



from __future__ import annotations

from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from typing import Any, List, Type

import requests

try:
    from bs4 import BeautifulSoup
except ImportError:
    raise ImportError(
        "Could not import scraper dependent packages. "
        "Please install it with `pip install beautifulsoup4 lxml`."
    )
from langchain_community.document_loaders import PyMuPDFLoader
from langchain_community.retrievers import ArxivRetriever
from pydantic import Field

from erniebot_agent.tools.base import Tool, ToolParameterView


class ScraperToolInputItemView(ToolParameterView):
    url: str = Field(description="http 有效链接")


class ScraperToolInputView(ToolParameterView):
    urls: List[ScraperToolInputItemView] = Field(description="http 有效链接列表")


class ScraperToolOutputView(ToolParameterView):
    result: List[str] = Field(description="每个 URL 链接中有效的内容")


class ScraperTool(Tool):
    """
    Scraper class to extract the content from the links
    """

    description: str = "能够从 HTTP URL 链接中提取文本内容"
    input_type: Type[ToolParameterView] = ScraperToolInputView
    output_type: Type[ToolParameterView] = ScraperToolOutputView

    def __init__(self, user_agent):
        """
        Initialize the Scraper class.
        Args:
            urls:
        """
        self.session = requests.Session()
        self.session.headers.update({"User-Agent": user_agent})

    async def __call__(self, urls: List[dict]) -> Any:
        urls = [item["url"] for item in urls]
        partial_extract = partial(self.extract_data_from_link, session=self.session)
        with ThreadPoolExecutor(max_workers=20) as executor:
            contents = executor.map(partial_extract, urls)
        res = [content for content in contents if content["raw_content"] is not None]
        return {"result": res}

    def extract_data_from_link(self, link, session):
        """
        Extracts the data from the link
        """
        content = ""
        try:
            if link.endswith(".pdf"):
                content = self.scrape_pdf_with_pymupdf(link)
            elif "arxiv.org" in link:
                doc_num = link.split("/")[-1]
                content = self.scrape_pdf_with_arxiv(doc_num)
            elif link:
                content = self.scrape_text_with_bs(link, session)

            if len(content) < 100:
                return {"url": link, "raw_content": None}
            return {"url": link, "raw_content": content}
        except Exception:
            return {"url": link, "raw_content": None}

    def scrape_text_with_bs(self, link, session):
        response = session.get(link, timeout=4)
        soup = BeautifulSoup(response.content, "lxml", from_encoding=response.encoding)

        for script_or_style in soup(["script", "style"]):
            script_or_style.extract()

        raw_content = self.get_content_from_url(soup)
        lines = (line.strip() for line in raw_content.splitlines())
        chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
        content = "\n".join(chunk for chunk in chunks if chunk)
        return content

    def scrape_pdf_with_pymupdf(self, url) -> str:
        """Scrape a pdf with pymupdf

        Args:
            url (str): The url of the pdf to scrape

        Returns:
            str: The text scraped from the pdf
        """
        loader = PyMuPDFLoader(url)
        doc = loader.load()
        return str(doc)

    def scrape_pdf_with_arxiv(self, query) -> str:
        """Scrape a pdf with arxiv
        default document length of 70000 about ~15 pages or None for no limit

        Args:
            query (str): The query to search for

        Returns:
            str: The text scraped from the pdf
        """
        retriever = ArxivRetriever(load_max_docs=2, doc_content_chars_max=None)  # type: ignore
        docs = retriever.get_relevant_documents(query=query)
        return docs[0].page_content

    def get_content_from_url(self, soup):
        """Get the text from the soup

        Args:
            soup (BeautifulSoup): The soup to get the text from

        Returns:
            str: The text from the soup
        """
        text = ""
        tags = ["p", "pre", "h1", "h2", "h3", "h4", "h5"]
        for element in soup.find_all(tags):  # Find all the <p> elements
            text += element.text + "\n\n"

        return text



# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import logging
from typing import Dict, List, Optional, Sequence, TypedDict

import erniebot.utils.token_helper as token_helper
from typing_extensions import Self

from erniebot_agent.file import File, protocol
from erniebot_agent.file.remote_file import RemoteFile

_logger = logging.getLogger(__name__)


class FunctionCall(TypedDict):
    name: str
    thoughts: str
    arguments: str


class SearchInfo(TypedDict):
    results: List[Dict]


class TokenUsage(TypedDict):
    prompt_tokens: int
    completion_tokens: int


class PluginInfo(Dict):
    names: List[str]


class Message:
    """
    Base class of the message.

    Args:
        role (str): character of the message.
        content (str): content of the message.
        token_count (Optional[int], optional): number of tokens of the message content. Defaults to None.

    Attributes:
        role (str): character of the message.
        content (str): content of the message.
        token_count (Optional[int]): number of tokens of the message content.

    Examples:
        >>> Message("user", "hello")
        <role: user, content: hello>
        >>> Message("user", "hello", token_count=5)
        <role: user, content: hello, token_count: 5>

    """

    def __init__(self, role: str, content: str, token_count: Optional[int] = None):
        self._role = role
        self._content = content
        self._token_count = token_count
        self._to_dict_keys = ["role", "content"]

    @property
    def role(self) -> str:
        return self._role

    @property
    def content(self) -> str:
        return self._content

    @property
    def token_count(self):
        if self._token_count is None:
            raise AttributeError("The token count of the message has not been set.")
        return self._token_count

    @token_count.setter
    def token_count(self, token_count: int):
        if self._token_count is not None:
            _logger.warning("The token count of the message has been set before")
        self._token_count = token_count

    def to_dict(self) -> Dict[str, str]:
        """
        Transfer the message to a dict, which is used to chat with models by ERNIB Bot SDK.
        """
        res = {}
        for name in self._to_dict_keys:
            res[name] = getattr(self, name)
        return res

    def __str__(self) -> str:
        return f"<{self._get_attrs_str()}>"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._get_attrs_str()}>"

    def _get_attrs_str(self) -> str:
        parts: List[str] = []
        for name in self._to_dict_keys:
            value = getattr(self, name)
            if value is not None and value != "":
                parts.append(f"{name}: {repr(value)}")
        if self._token_count is not None:
            parts.append(f"token_count: {self._token_count}")
        return ", ".join(parts)


class SystemMessage(Message):
    """
    Definition of system message, such that the feature of the Agent can be customized.

    Args:
        content (str): the content of the message.

    Attributes:
        role (str): character of the message.
        content (str): content of the message.
        token_count (Optional[int]): number of tokens of the message content.

    Examples:
        >>> from erniebot_agent.messages import SystemMessage
        >>> SystemMessage("you are an assistant useful for ocr.")
        <role: system, content: you are an assistant useful for ocr.>
        >>> SystemMessage("you are an assistant useful for ocr.").to_dict()
        {'role': 'system', 'content': 'you are an assistant useful for ocr.'}
        >>> SystemMessage("you are an assistant useful for ocr.").token_count
        2
        >>> SystemMessage("you are an assistant useful for ocr.").token_count = 3
        >>> SystemMessage("you are an assistant useful for ocr.").token_count
        3

    """

    def __init__(self, content: str):
        super().__init__(role="system", content=content, token_count=len(content))


class HumanMessage(Message):
    """
    The definition of the message created by a human.

    Args:
        content (str): the content of the message.

    Attributes:
        role (str): character of the message.
        content (str): content of the message.
        token_count (Optional[int]): number of tokens of the message content.

    Examples:
        >>> from erniebot_agent.messages import HumanMessage
        >>> HumanMessage("I want to order a pizza.")
        <role: user, content: I want to order a pizza.>
        >>> prompt = "What is the text in this image?"
        >>> files = [await file_manager.create_file_from_path(file_path="ocr_img.jpg", file_type="remote")]
        >>> message = await HumanMessage.create_with_files(
                prompt, files, include_file_urls=True)
        >>> message
        <role: user, content: W h ha t.<file>File-local-xxxx</file><url>{url}</url>.>

    """

    def __init__(self, content: str):
        super().__init__(role="user", content=content)

    @classmethod
    async def create_with_files(
        cls, text: str, files: Sequence[File], *, include_file_urls: bool = False
    ) -> Self:
        """
        create a Human Message with file input

        Args:
            text: content of the message.
            files (Sequence[File]): The files that the message contains.
            include_file_urls: Whehter to include file URLs in the content of message.

        Returns:
            A HumanMessage object that contains file in the content.

        Raises:
            RuntimeError: Only `RemoteFile` objects can set include_file_urls as True.
        """

        def _get_file_reprs(files: Sequence[File]) -> List[str]:
            file_reprs: List[str] = []
            for file in files:
                file_reprs.append(file.get_file_repr())
            return file_reprs

        async def _create_file_reprs_with_urls(files: Sequence[File]) -> List[str]:
            file_reprs = []
            for file in files:
                if not isinstance(file, RemoteFile):
                    raise TypeError("Only `RemoteFile` objects can have URLs in their representations.")
                url = await file.create_temporary_url()
                file_reprs.append(file.get_file_repr_with_url(url))

            return file_reprs

        def _append_files_repr_to_text(text: str, files_repr: str) -> str:
            return f"{text}\n{files_repr}"

        if len(files) > 0:
            if len(protocol.extract_file_ids(text)) > 0:
                _logger.warning("File IDs were found in the text. The provided files will be ignored.")
            else:
                if include_file_urls:
                    file_reprs = await _create_file_reprs_with_urls(files)
                else:
                    file_reprs = _get_file_reprs(files)
                files_repr = "\n".join(file_reprs)
                content = _append_files_repr_to_text(text, files_repr)
        else:
            content = text
        return cls(content)


class AIMessage(Message):
    """
    The definition of the message from an assistant.

    Args:
        content (str): the content of the message.
        function_call (Optional[FunctionCall], optional): The function that agent calls. Defaults to None.
        token_usage (Optional[TokenUsage], optional): the token usage calculate by ERNIE. Defaults to None.
        search_info (Optional[SearchInfo], optional):
                The SearchInfo content of the chat model's response. Defaults to None.

    Attributes:
        role (str): character of the message.
        content (str): content of the message.
        token_count (Optional[int]): number of tokens of the message content.
        function_call (Optional[FunctionCall]): The function that agent calls.
        query_tokens_count (int): the number of tokens in the query.
        search_info (Optional[SearchInfo]): The SearchInfo in the chat model's response.

    Examples:
        >>> human_message = HumanMessage(content="What is the text in this image?")
        >>> ai_message = AIMessage(
            function_call={"name": "OCR", "thoughts": "The user want to know the text in the image,
                    I need to use the OCR tool",
                    "arguments": "{\"imgae_byte_str\": file-remote-xxxx, \"lang\": "en"}"},
            token_usage={"prompt_tokens": 10, "completion_tokens": 20},
            search_info={}]}
            )
        >>> human_message.content
        "What is the text in this image?"
        >>> ai_message.function_call
        {"name": "OCR",
            "thoughts": "The user want to know the text in the image, I need to use the OCR tool",
            "arguments": "{\"imgae_byte_str\": file-remote-xxxx, \"lang\": "en"}"}
    """

    def __init__(
        self,
        content: str,
        function_call: Optional[FunctionCall] = None,
        token_usage: Optional[TokenUsage] = None,
        plugin_info: Optional[PluginInfo] = None,
        search_info: Optional[SearchInfo] = None,
        clarify: Optional[bool] = False,
    ):
        if token_usage is None:
            prompt_tokens = 0
            completion_tokens = token_helper.approx_num_tokens(content)
        else:
            prompt_tokens, completion_tokens = self._parse_token_count(token_usage)
        super().__init__(role="assistant", content=content, token_count=completion_tokens)
        self.function_call = function_call
        self.query_tokens_count = prompt_tokens
        self.plugin_info = plugin_info
        self.search_info = search_info
        self.clarify = clarify
        self._to_dict_keys = ["role", "content", "function_call", "plugin_info", "search_info"]

    def _parse_token_count(self, token_usage: TokenUsage):
        """Parse the token count information from LLM."""
        return token_usage["prompt_tokens"], token_usage["completion_tokens"]


class FunctionMessage(Message):
    """
    The definition of a message that calls tools, containing the result of a function call.

    Args:
        name (str): the name of the function.
        content (str): the content of the message.

    Attributes:
        name (str): the name of the function.
        role (str): character of the message.
        content (str): content of the message.
        token_count (Optional[int]): number of tokens of the message content.

    Examples:
        >>> function_message = FunctionMessage(name="OCR", content="The text in the image is: 1234567")
        >>> function_message.name
        "OCR"
        >>> function_message.content
        "The text in the image is: 1234567"
        >>> function_message.role
        "function"
        >>> function_message.token_count
        0
    """

    def __init__(self, name: str, content: str):
        super().__init__(role="function", content=content)
        self.name = name
        self._to_dict_keys = ["role", "name", "content"]


class AIMessageChunk(AIMessage):
    """The definition of a message chunk from an assistant."""



import abc
import json
import logging
from typing import (
    Any,
    Dict,
    Final,
    Iterable,
    List,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Union,
    final,
)

from erniebot_agent.agents.base import BaseAgent
from erniebot_agent.agents.callback.callback_manager import CallbackManager
from erniebot_agent.agents.callback.default import get_default_callbacks
from erniebot_agent.agents.callback.handlers.base import CallbackHandler
from erniebot_agent.agents.mixins import GradioMixin
from erniebot_agent.agents.schema import AgentResponse, LLMResponse, ToolResponse
from erniebot_agent.chat_models.erniebot import BaseERNIEBot
from erniebot_agent.file import (
    File,
    FileManager,
    GlobalFileManagerHandler,
    get_default_file_manager,
)
from erniebot_agent.memory import Memory, WholeMemory
from erniebot_agent.memory.messages import Message, SystemMessage
from erniebot_agent.tools.base import BaseTool
from erniebot_agent.tools.tool_manager import ToolManager
from erniebot_agent.utils.exceptions import FileError

_PLUGINS_WO_FILE_IO: Final[Tuple[str]] = ("eChart",)

_logger = logging.getLogger(__name__)


class Agent(GradioMixin, BaseAgent[BaseERNIEBot]):
    """The base class for agents.

    Typically, this class should be the base class for custom agent classes. A
    class derived from this class must implement how the agent orchestates the
    components to complete tasks.

    Attributes:
        llm: The LLM that the agent uses.
        memory: The message storage that keeps the chat history.
    """

    llm: BaseERNIEBot
    memory: Memory

    def __init__(
        self,
        llm: BaseERNIEBot,
        tools: Union[ToolManager, Iterable[BaseTool]],
        *,
        memory: Optional[Memory] = None,
        system: Optional[str] = None,
        callbacks: Optional[Union[CallbackManager, Iterable[CallbackHandler]]] = None,
        file_manager: Optional[FileManager] = None,
        plugins: Optional[List[str]] = None,
    ) -> None:
        """Initialize an agent.

        Args:
            llm: An LLM for the agent to use.
            tools: Tools for the agent to use.
            memory: A memory object that equips the agent to remember chat
                history. If not specified, a new WholeMemory object will be instantiated.
            system: A message that tells the LLM how to interpret the
                conversations.
            callbacks: A list of callback handlers for the agent to use. If
                `None`, a default list of callbacks will be used.
            file_manager: A file manager for the agent to interact with files.
                If `None`, a global file manager that can be shared among
                different components will be implicitly created and used.
            plugins: A list of names of the plugins for the agent to use. If
                `None`, the agent will use a default list of plugins. Set
                `plugins` to `[]` to disable the use of plugins.
        """
        super().__init__()
        self.llm = llm
        if isinstance(tools, ToolManager):
            self._tool_manager = tools
        else:
            self._tool_manager = ToolManager(tools)
        if memory is None:
            memory = self._create_default_memory()
        self.memory = memory

        self.system = SystemMessage(system) if system is not None else system
        if self.system is not None:
            self.memory.set_system_message(self.system)

        if callbacks is None:
            callbacks = get_default_callbacks()
        if isinstance(callbacks, CallbackManager):
            self._callback_manager = callbacks
        else:
            self._callback_manager = CallbackManager(callbacks)
        self._file_manager = file_manager or get_default_file_manager()
        self._plugins = plugins
        self._init_file_needs_url()

    @final
    async def run(self, prompt: str, files: Optional[Sequence[File]] = None) -> AgentResponse:
        """Run the agent asynchronously.

        Args:
            prompt: A natural language text describing the task that the agent
                should perform.
            files: A list of files that the agent can use to perform the task.

        Returns:
            Response from the agent.
        """
        if files:
            await self._ensure_managed_files(files)
        await self._callback_manager.on_run_start(agent=self, prompt=prompt)
        try:
            agent_resp = await self._run(prompt, files)
        except BaseException as e:
            await self._callback_manager.on_run_error(agent=self, error=e)
            raise e
        else:
            await self._callback_manager.on_run_end(agent=self, response=agent_resp)
        return agent_resp

    @final
    async def run_llm(
        self,
        messages: List[Message],
        **llm_opts: Any,
    ) -> LLMResponse:
        """Run the LLM asynchronously.

        Args:
            messages: The input messages.
            llm_opts: Options to pass to the LLM.

        Returns:
            Response from the LLM.
        """
        await self._callback_manager.on_llm_start(agent=self, llm=self.llm, messages=messages)
        try:
            llm_resp = await self._run_llm(messages, **(llm_opts or {}))
        except (Exception, KeyboardInterrupt) as e:
            await self._callback_manager.on_llm_error(agent=self, llm=self.llm, error=e)
            raise e
        else:
            await self._callback_manager.on_llm_end(agent=self, llm=self.llm, response=llm_resp)
        return llm_resp

    @final
    async def run_tool(self, tool_name: str, tool_args: str) -> ToolResponse:
        """Run the specified tool asynchronously.

        Args:
            tool_name: The name of the tool to run.
            tool_args: The tool arguments in JSON format.

        Returns:
            Response from the tool.
        """
        tool = self._tool_manager.get_tool(tool_name)
        await self._callback_manager.on_tool_start(agent=self, tool=tool, input_args=tool_args)
        try:
            tool_resp = await self._run_tool(tool, tool_args)
        except (Exception, KeyboardInterrupt) as e:
            await self._callback_manager.on_tool_error(agent=self, tool=tool, error=e)
            raise e
        else:
            await self._callback_manager.on_tool_end(agent=self, tool=tool, response=tool_resp)
        return tool_resp

    def load_tool(self, tool: BaseTool) -> None:
        """Load a tool into the agent.

        Args:
            tool: The tool to load.
        """
        self._tool_manager.add_tool(tool)

    def unload_tool(self, tool: Union[BaseTool, str]) -> None:
        """Unload a tool from the agent.

        Args:
            tool: The tool to unload.
        """
        if isinstance(tool, str):
            tool = self.get_tool(tool)
        self._tool_manager.remove_tool(tool)

    def get_tool(self, tool_name: str) -> BaseTool:
        """Get a tool by name."""
        return self._tool_manager.get_tool(tool_name)

    def get_tools(self) -> List[BaseTool]:
        """Get the tools that the agent can choose from."""
        return self._tool_manager.get_tools()

    def reset_memory(self) -> None:
        """Clear the chat history."""
        self.memory.clear_chat_history()

    def get_file_manager(self) -> FileManager:
        # Can we create a lazy proxy for the global file manager and simply set
        # and use `self._file_manager`?
        if self._file_manager is None:
            file_manager = GlobalFileManagerHandler().get()
        else:
            file_manager = self._file_manager
        return file_manager

    @abc.abstractmethod
    async def _run(self, prompt: str, files: Optional[Sequence[File]] = None) -> AgentResponse:
        raise NotImplementedError

    async def _run_llm(self, messages: List[Message], **opts: Any) -> LLMResponse:
        for reserved_opt in ("stream", "system", "plugins"):
            if reserved_opt in opts:
                raise TypeError(f"`{reserved_opt}` should not be set.")

        if "functions" not in opts:
            functions = self._tool_manager.get_tool_schemas()
        else:
            functions = opts.pop("functions")

        if hasattr(self.llm, "system"):
            _logger.warning(
                "The `system` message has already been set in the agent;"
                "the `system` message configured in ERNIEBot will become ineffective."
            )
        opts["system"] = self.system.content if self.system is not None else None
        opts["plugins"] = self._plugins
        llm_ret = await self.llm.chat(messages, stream=False, functions=functions, **opts)
        return LLMResponse(message=llm_ret)

    async def _run_tool(self, tool: BaseTool, tool_args: str) -> ToolResponse:
        parsed_tool_args = self._parse_tool_args(tool_args)
        file_manager = self.get_file_manager()
        # XXX: Sniffing is less efficient and probably unnecessary.
        # Can we make a protocol to statically recognize file inputs and outputs
        # or can we have the tools introspect about this?
        input_files = file_manager.sniff_and_extract_files_from_dict(parsed_tool_args)
        tool_ret = await tool(**parsed_tool_args)
        if isinstance(tool_ret, dict):
            output_files = file_manager.sniff_and_extract_files_from_dict(tool_ret)
        else:
            output_files = []
        tool_ret_json = json.dumps(tool_ret, ensure_ascii=False)
        return ToolResponse(json=tool_ret_json, input_files=input_files, output_files=output_files)

    def _create_default_memory(self) -> Memory:
        return WholeMemory()

    def _init_file_needs_url(self):
        self.file_needs_url = False
        if self._plugins:
            for plugin in self._plugins:
                if plugin not in _PLUGINS_WO_FILE_IO:
                    self.file_needs_url = True

    def _parse_tool_args(self, tool_args: str) -> Dict[str, Any]:
        try:
            args_dict = json.loads(tool_args)
        except json.JSONDecodeError:
            raise ValueError(f"`tool_args` cannot be parsed as JSON. `tool_args`: {tool_args}")

        if not isinstance(args_dict, dict):
            raise ValueError(f"`tool_args` cannot be interpreted as a dict. `tool_args`: {tool_args}")
        return args_dict

    async def _ensure_managed_files(self, files: Sequence[File]) -> None:
        def _raise_exception(file: File) -> NoReturn:
            raise FileError(f"{repr(file)} is not managed by the file manager of the agent.")

        file_manager = self.get_file_manager()
        for file in files:
            try:
                managed_file = file_manager.look_up_file_by_id(file.id)
            except FileError:
                _raise_exception(file)
            if file is not managed_file:
                _raise_exception(file)



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import uuid
from typing import Any, Dict

import anyio

from erniebot_agent.file import protocol
from erniebot_agent.file.base import BaseFile


def create_local_file_from_path(
    file_path: pathlib.Path,
    file_purpose: protocol.FilePurpose,
    file_metadata: Dict[str, Any],
) -> "LocalFile":
    """
    Create a LocalFile object from a local file path.

    Args:
        file_path (pathlib.Path): The path to the local file.
        file_purpose (protocol.FilePurpose): The purpose or use case of the file,
                                             including "assistants" and "assistants_output".
        file_metadata (Dict[str, Any]): Additional metadata associated with the file.

    Returns:
        LocalFile: The created LocalFile object.

    Raises:
        FileNotFoundError: If the specified file does not exist.

    """
    if not file_path.exists():
        raise FileNotFoundError(f"File {file_path} does not exist.")
    file_id = _generate_local_file_id()
    filename = file_path.name
    byte_size = file_path.stat().st_size
    created_at = protocol.get_timestamp()
    file = LocalFile(
        id=file_id,
        filename=filename,
        byte_size=byte_size,
        created_at=created_at,
        purpose=file_purpose,
        metadata=file_metadata,
        path=file_path,
    )
    return file


class LocalFile(BaseFile):
    """
    Represents a local file.

    Attributes:
        id (str): Unique identifier for the file.
        filename (str): File name.
        byte_size (int): Size of the file in bytes.
        created_at (str): Timestamp indicating the file creation time.
        purpose (str): Purpose or use case of the file,
                       including "assistants" and "assistants_output".
        metadata (Dict[str, Any]): Additional metadata associated with the file.
        path (pathlib.Path): The path to the local file.

    Methods:
        read_contents: Asynchronously read the contents of the local file.
        write_contents_to: Asynchronously write the file contents to a local path.
        get_file_repr: Return a string representation for use in specific contexts.

    """

    def __init__(
        self,
        *,
        id: str,
        filename: str,
        byte_size: int,
        created_at: str,
        purpose: protocol.FilePurpose,
        metadata: Dict[str, Any],
        path: pathlib.Path,
        validate_file_id: bool = True,
    ) -> None:
        """
        Initialize a LocalFile object.

        Args:
            id (str): The unique identifier for the file.
            filename (str): The name of the file.
            byte_size (int): The size of the file in bytes.
            created_at (str): The timestamp indicating the file creation time.
            purpose (protocol.FilePurpose): The purpose or use case of the file.
            metadata (Dict[str, Any]): Additional metadata associated with the file.
            path (pathlib.Path): The path to the local file.
            validate_file_id (bool): Flag to validate the file ID. Default is True.

        Raises:
            ValueError: If the file ID is invalid.

        """
        if validate_file_id:
            if not protocol.is_local_file_id(id):
                raise ValueError(f"Invalid file ID: {id}")
        if not protocol.is_valid_file_purpose(purpose):
            raise ValueError(f"Invalid file purpose: {purpose}")
        super().__init__(
            id=id,
            filename=filename,
            byte_size=byte_size,
            created_at=created_at,
            purpose=purpose,
            metadata=metadata,
        )
        self.path = path

    async def read_contents(self) -> bytes:
        """Asynchronously read the contents of the local file."""
        return await anyio.Path(self.path).read_bytes()

    def _get_attrs_str(self) -> str:
        attrs_str = super()._get_attrs_str()
        attrs_str += f", path: {repr(self.path)}"
        return attrs_str


def _generate_local_file_id():
    return protocol.create_local_file_id_from_uuid(str(uuid.uuid1()))



from __future__ import annotations

import json
import logging
import os
import tempfile
from dataclasses import asdict, dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Type

import requests
from openapi_spec_validator.readers import read_from_filename
from yaml import safe_dump

from erniebot_agent.file import FileManager
from erniebot_agent.memory.messages import (
    AIMessage,
    FunctionCall,
    HumanMessage,
    Message,
)
from erniebot_agent.tools.remote_tool import RemoteTool, tool_registor
from erniebot_agent.tools.schema import (
    Endpoint,
    EndpointInfo,
    RemoteToolView,
    ToolParameterView,
    scrub_dict,
)
from erniebot_agent.tools.utils import validate_openapi_yaml
from erniebot_agent.utils import config_from_environ as C
from erniebot_agent.utils.exceptions import RemoteToolError
from erniebot_agent.utils.http import url_file_exists

_logger = logging.getLogger(__name__)


@dataclass
class RemoteToolkit:
    """RemoteToolkit can be converted by openapi.yaml and endpoint"""

    openapi: str
    info: EndpointInfo
    servers: List[Endpoint]
    paths: List[RemoteToolView]
    file_manager: Optional[FileManager]

    component_schemas: dict[str, Type[ToolParameterView]]
    headers: dict
    examples: List[Message] = field(default_factory=list)
    _AISTUDIO_HUB_BASE_URL: ClassVar[str] = "https://aistudio-hub.baidu.com"

    @property
    def tool_name_prefix(self) -> str:
        return f"{self.info.title}/{self.info.version}"

    def __getitem__(self, tool_name: str) -> RemoteTool:
        return self.get_tool(tool_name)

    def get_tools(self) -> List[RemoteTool]:
        TOOL_CLASS = tool_registor.get_tool_class(self.info.title)
        return [
            TOOL_CLASS(
                path,
                self.servers[0].url,
                self.headers,
                self.info.version,
                file_manager=self.file_manager,
                examples=self.get_examples_by_name(path.name),
                tool_name_prefix=self.tool_name_prefix,
            )
            for path in self.paths
        ]

    def get_examples_by_name(self, tool_name: str) -> List[Message]:
        """get examples by tool-name

        Args:
            tool_name (str): the name of the tool

        Returns:
            List[Message]: the messages
        """
        # 1. split messages
        tool_examples: List[List[Message]] = []
        examples: List[Message] = []
        for example in self.examples:
            if isinstance(example, HumanMessage):
                if len(examples) == 0:
                    examples.append(example)
                else:
                    tool_examples.append(examples)
                    examples = [example]
            else:
                examples.append(example)

        if len(examples) > 0:
            tool_examples.append(examples)

        final_exampels: List[Message] = []
        # 2. find the target tool examples or empty messages
        for examples in tool_examples:
            tool_names = [
                example.function_call.get("name", None)
                for example in examples
                if isinstance(example, AIMessage) and example.function_call is not None
            ]
            tool_names = [name for name in tool_names if name]

            if tool_name in tool_names:
                # 3. prepend `tool_name_prefix` to all tool names in examples
                for example in examples:
                    if isinstance(example, AIMessage) and example.function_call is not None:
                        original_tool_name = example.function_call["name"]
                        example.function_call["name"] = f"{self.tool_name_prefix}/{original_tool_name}"
                final_exampels.extend(examples)

        return final_exampels

    def get_tool(self, tool_name: str) -> RemoteTool:
        paths = [path for path in self.paths if path.name == tool_name]
        if len(paths) == 0:
            raise RemoteToolError(
                f"`{tool_name}` not found under RemoteToolkit `{self.tool_name_prefix}`", stage="Loading"
            )
        elif len(paths) > 1:
            raise RemoteToolError(
                f"Found duplicate `{tool_name}` under RemoteToolkit `{self.tool_name_prefix}`",
                stage="Loading",
            )

        TOOL_CLASS = tool_registor.get_tool_class(self.info.title)
        return TOOL_CLASS(
            paths[0],
            self.servers[0].url,
            self.headers,
            self.info.version,
            file_manager=self.file_manager,
            examples=self.get_examples_by_name(tool_name),
            tool_name_prefix=self.tool_name_prefix,
        )

    def to_openapi_dict(self) -> dict:
        """convert plugin schema to openapi spec dict"""
        spec_dict = {
            "openapi": self.openapi,
            "info": asdict(self.info),
            "servers": [asdict(server) for server in self.servers],
            "paths": {tool_view.uri: tool_view.to_openapi_dict() for tool_view in self.paths},
            "components": {
                "schemas": {
                    uri: parameters_view.to_openapi_dict()
                    for uri, parameters_view in self.component_schemas.items()
                }
            },
        }
        return scrub_dict(spec_dict, remove_empty_dict=True) or {}

    def to_openapi_file(self, file: str):
        """generate openapi configuration file

        Args:
            file (str): the path of the openapi yaml file
        """
        spec_dict = self.to_openapi_dict()
        with open(file, "w+", encoding="utf-8") as f:
            safe_dump(spec_dict, f, indent=4)

    @classmethod
    def from_openapi_dict(
        cls,
        openapi_dict: Dict[str, Any],
        access_token: Optional[str] = None,
        file_manager: Optional[FileManager] = None,
    ) -> RemoteToolkit:
        info = EndpointInfo(**openapi_dict["info"])
        servers = [Endpoint(**server) for server in openapi_dict.get("servers", [])]

        if access_token is None:
            access_token = C.get_global_access_token()

        # components
        component_schemas = openapi_dict["components"]["schemas"]
        fields = {}
        for schema_name, schema in component_schemas.items():
            parameter_view = ToolParameterView.from_openapi_dict(schema)
            fields[schema_name] = parameter_view

        # paths
        paths = []
        for path, path_info in openapi_dict.get("paths", {}).items():
            if ".well-known/openapi.yaml" in path:
                continue

            for method, path_method_info in path_info.items():
                paths.append(
                    RemoteToolView.from_openapi_dict(
                        uri=path,
                        method=method,
                        version=info.version,
                        path_info=path_method_info,
                        parameters_views=fields,
                    )
                )

        return RemoteToolkit(
            openapi=openapi_dict["openapi"],
            info=info,
            servers=servers,
            paths=paths,
            component_schemas=fields,
            headers=cls._get_authorization_headers(access_token),
            file_manager=file_manager,
        )

    @classmethod
    def from_openapi_file(
        cls, file: str, access_token: Optional[str] = None, file_manager: Optional[FileManager] = None
    ) -> RemoteToolkit:
        """only support openapi v3.0.1

        Args:
            file (str): the path of openapi yaml file
            access_token (Optional[str]): the path of openapi yaml file
        """
        if not validate_openapi_yaml(file):
            raise RemoteToolError(f"invalid openapi yaml file: {file}", stage="Loading")

        if access_token is None:
            access_token = C.get_global_access_token()

        spec_dict, _ = read_from_filename(file)
        return cls.from_openapi_dict(
            spec_dict, access_token=access_token, file_manager=file_manager  # type: ignore
        )

    @classmethod
    def _get_authorization_headers(cls, access_token: Optional[str]) -> dict:
        headers = {"Content-Type": "application/json"}
        if access_token is None:
            _logger.warning("access_token is NOT provided, this may cause 403 HTTP error..")
        else:
            headers["Authorization"] = f"token {access_token}"
        return headers

    @classmethod
    def from_aistudio(
        cls,
        tool_id: str,
        version: Optional[str] = None,
        access_token: Optional[str] = None,
        file_manager: Optional[FileManager] = None,
    ) -> RemoteToolkit:
        from urllib.parse import urlparse

        if access_token is None:
            access_token = C.get_global_access_token()

        aistudio_base_url = os.getenv("AISTUDIO_HUB_BASE_URL", cls._AISTUDIO_HUB_BASE_URL)
        parsed_url = urlparse(aistudio_base_url)
        tool_url = parsed_url._replace(netloc=f"tool-{tool_id}.{parsed_url.netloc}").geturl()
        return cls.from_url(tool_url, version=version, access_token=access_token, file_manager=file_manager)

    @classmethod
    def from_url(
        cls,
        url: str,
        version: Optional[str] = None,
        access_token: Optional[str] = None,
        file_manager: Optional[FileManager] = None,
    ) -> RemoteToolkit:
        # 1. download openapy.yaml file to temp directory
        if not url.endswith("/"):
            url += "/"
        openapi_yaml_url = url + ".well-known/openapi.yaml"

        if version:
            openapi_yaml_url = openapi_yaml_url + "?version=" + version

        if access_token is None:
            access_token = C.get_global_access_token()

        with tempfile.TemporaryDirectory() as temp_dir:
            response = requests.get(openapi_yaml_url, headers=cls._get_authorization_headers(access_token))
            if response.status_code != 200:
                _logger.debug(f"The resource requested returned the following headers: {response.headers}")
                raise RemoteToolError(
                    f"`{openapi_yaml_url}` returned {response.status_code}: {response.text}", stage="Loading"
                )

            file_content = response.content.decode("utf-8")
            if not file_content.strip():
                raise RemoteToolError(f"the content is empty from: {openapi_yaml_url}", stage="Loading")

            file_path = os.path.join(temp_dir, "openapi.yaml")
            with open(file_path, "w+", encoding="utf-8") as f:
                f.write(file_content)

            spec_dict: dict = read_from_filename(file_path)[0]  # type: ignore

            url = url.strip("/")

            if "servers" not in spec_dict:
                spec_dict["servers"] = [{"url": url}]

            toolkit = RemoteToolkit.from_openapi_dict(
                spec_dict, access_token=access_token, file_manager=file_manager
            )
            for server in toolkit.servers:
                server.url = url

            toolkit.examples = cls.load_remote_examples_yaml(url, access_token)

        return toolkit

    @classmethod
    def load_remote_examples_yaml(cls, url: str, access_token: Optional[str] = None) -> List[Message]:
        """load remote examples by url: url/.well-known/examples.yaml

        Args:
            url (str): the base url of the remote toolkit
        """
        if not url.endswith("/"):
            url += "/"
        examples_yaml_url = url + ".well-known/examples.yaml"
        if not url_file_exists(examples_yaml_url, cls._get_authorization_headers(access_token)):
            return []

        if access_token is None:
            access_token = C.get_global_access_token()

        examples = []
        with tempfile.TemporaryDirectory() as temp_dir:
            response = requests.get(examples_yaml_url, headers=cls._get_authorization_headers(access_token))
            if response.status_code != 200:
                _logger.debug(f"The resource requested returned the following headers: {response.headers}")
                raise RemoteToolError(
                    f"`{examples_yaml_url}` returned {response.status_code}: {response.text}",
                    stage="Loading",
                )

            file_content = response.content.decode("utf-8")
            if not file_content.strip():
                raise RemoteToolError(f"the content is empty from: {examples_yaml_url}", stage="Loading")

            file_path = os.path.join(temp_dir, "examples.yaml")
            with open(file_path, "w+", encoding="utf-8") as f:
                f.write(file_content)

            examples = cls.load_examples_yaml(file_path)

        return examples

    @classmethod
    def load_examples_dict(cls, examples_dict: Dict[str, Any]) -> List[Message]:
        messages: List[Message] = []
        for examples in examples_dict["examples"]:
            examples = examples["context"]
            for example in examples:
                if "user" == example["role"]:
                    messages.append(HumanMessage(example["content"]))
                elif "bot" in example["role"]:
                    plugin = example["plugin"]
                    if "operationId" in plugin:
                        function_call: FunctionCall = {
                            "name": plugin["operationId"],
                            "thoughts": plugin["thoughts"],
                            "arguments": json.dumps(plugin["requestArguments"], ensure_ascii=False),
                        }
                    else:
                        function_call = {
                            "name": "",
                            "thoughts": plugin["thoughts"],
                            "arguments": "{}",
                        }  # type: ignore
                    messages.append(AIMessage("", function_call=function_call))
                else:
                    raise RemoteToolError(f"invald role: <{example['role']}>", stage="Loading")
        return messages

    @classmethod
    def load_examples_yaml(cls, file: str) -> List[Message]:
        """load examples from yaml file

        Args:
            file (str): the path of examples file

        Returns:
            List[Message]: the list of messages
        """
        content: dict = read_from_filename(file)[0]  # type: ignore
        if len(content) == 0 or "examples" not in content:
            raise RemoteToolError("invalid examples configuration file", stage="Loading")
        return cls.load_examples_dict(content)

    def function_call_schemas(self) -> List[dict]:
        return [tool.function_call_schema() for tool in self.get_tools()]



from __future__ import annotations

from typing import Any, AsyncIterator, Dict, Iterator, List, Mapping, Optional

from langchain.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain.chat_models.base import BaseChatModel
from langchain.pydantic_v1 import Field, root_validator
from langchain.schema import ChatGeneration, ChatResult
from langchain.schema.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    ChatMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage,
)
from langchain.schema.output import ChatGenerationChunk
from langchain.utils import get_from_dict_or_env

_MessageDict = Dict[str, Any]


class ErnieBotChat(BaseChatModel):
    """ERNIE Bot Chat large language models API.

    To use, you should have the ``erniebot`` python package installed, and the
    environment variable ``AISTUDIO_ACCESS_TOKEN`` set with your AI Studio access token.

    Example:
        .. code-block:: python
            from erniebot_agent.extensions.langchain.chat_models import ErnieBotChat
            erniebot_chat = ErnieBotChat(model="ernie-3.5")
    """

    client: Any = None
    aistudio_access_token: Optional[str] = None
    """AI Studio access token."""
    max_retries: int = 6
    """Maximum number of retries to make when generating."""
    streaming: bool = False
    """Whether to stream the results or not."""

    model: str = "ernie-3.5"
    """Model to use."""
    temperature: Optional[float] = 0.95
    """Sampling temperature to use."""
    top_p: Optional[float] = 0.8
    """Parameter of nucleus sampling that affects the diversity of generated content."""
    penalty_score: Optional[float] = 1
    """Penalty assigned to tokens that have been generated."""
    request_timeout: Optional[int] = 60
    """How many seconds to wait for the server to send data before giving up."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""

    ernie_client_id: Optional[str] = None
    ernie_client_secret: Optional[str] = None
    """For raising deprecation warnings."""

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling ERNIE Bot API."""
        normal_params = {
            "model": self.model,
            "top_p": self.top_p,
            "temperature": self.temperature,
            "penalty_score": self.penalty_score,
            "request_timeout": self.request_timeout,
        }
        return {**normal_params, **self.model_kwargs}

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        return self._default_params

    @property
    def _invocation_params(self) -> Dict[str, Any]:
        """Get the parameters used to invoke the model."""
        auth_cfg: Dict[str, Optional[str]] = {
            "api_type": "aistudio",
            "access_token": self.aistudio_access_token,
        }
        return {**{"_config_": {"max_retries": self.max_retries, **auth_cfg}}, **self._default_params}

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "erniebot"

    @root_validator()
    def validate_enviroment(cls, values: Dict) -> Dict:
        values["aistudio_access_token"] = get_from_dict_or_env(
            values,
            "aistudio_access_token",
            "AISTUDIO_ACCESS_TOKEN",
        )

        try:
            import erniebot

            values["client"] = erniebot.ChatCompletion
        except ImportError:
            raise ImportError(
                "Could not import erniebot python package. Please install it with `pip install erniebot`."
            )
        return values

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self.streaming:
            chunks = self._stream(messages, stop=stop, run_manager=run_manager, **kwargs)
            generation: Optional[ChatGenerationChunk] = None
            for chunk in chunks:
                if generation is None:
                    generation = chunk
                else:
                    generation += chunk
            assert generation is not None
            return ChatResult(generations=[generation])
        else:
            params = self._invocation_params
            params.update(kwargs)
            params["messages"] = self._convert_messages_to_dicts(messages)
            system_prompt = self._build_system_prompt_from_messages(messages)
            if system_prompt is not None:
                params["system"] = system_prompt
            params["stream"] = False
            response = self.client.create(**params)
            return self._build_chat_result_from_response(response)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self.streaming:
            chunks = self._astream(messages, stop=stop, run_manager=run_manager, **kwargs)
            generation: Optional[ChatGenerationChunk] = None
            async for chunk in chunks:
                if generation is None:
                    generation = chunk
                else:
                    generation += chunk
            assert generation is not None
            return ChatResult(generations=[generation])
        else:
            params = self._invocation_params
            params.update(kwargs)
            params["messages"] = self._convert_messages_to_dicts(messages)
            system_prompt = self._build_system_prompt_from_messages(messages)
            if system_prompt is not None:
                params["system"] = system_prompt
            params["stream"] = False
            response = await self.client.acreate(**params)
            return self._build_chat_result_from_response(response)

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        if stop is not None:
            raise ValueError("Currently, `stop` is not supported when streaming is enabled.")
        params = self._invocation_params
        params.update(kwargs)
        params["messages"] = self._convert_messages_to_dicts(messages)
        system_prompt = self._build_system_prompt_from_messages(messages)
        if system_prompt is not None:
            params["system"] = system_prompt
        params["stream"] = True
        for resp in self.client.create(**params):
            chunk = self._build_chunk_from_response(resp)
            yield chunk
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    async def _astream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        if stop is not None:
            raise ValueError("Currently, `stop` is not supported when streaming is enabled.")
        params = self._invocation_params
        params.update(kwargs)
        params["messages"] = self._convert_messages_to_dicts(messages)
        system_prompt = self._build_system_prompt_from_messages(messages)
        if system_prompt is not None:
            params["system"] = system_prompt
        params["stream"] = True
        async for resp in await self.client.acreate(**params):
            chunk = self._build_chunk_from_response(resp)
            yield chunk
            if run_manager:
                await run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    def _build_chat_result_from_response(self, response: Mapping[str, Any]) -> ChatResult:
        message_dict = self._build_dict_from_response(response)
        generation = ChatGeneration(
            message=self._convert_dict_to_message(message_dict),
            generation_info=dict(finish_reason="stop"),
        )
        token_usage = response.get("usage", {})
        llm_output = {"token_usage": token_usage, "model_name": self.model}
        return ChatResult(generations=[generation], llm_output=llm_output)

    def _build_chunk_from_response(self, response: Mapping[str, Any]) -> ChatGenerationChunk:
        message_dict = self._build_dict_from_response(response)
        message = self._convert_dict_to_message(message_dict)
        msg_chunk = AIMessageChunk(
            content=message.content,
            additional_kwargs=message.additional_kwargs,
        )
        return ChatGenerationChunk(message=msg_chunk)

    def _build_dict_from_response(self, response: Mapping[str, Any]) -> _MessageDict:
        message_dict: _MessageDict = {"role": "assistant"}
        if "function_call" in response:
            message_dict["content"] = None
            message_dict["function_call"] = response["function_call"]
        else:
            message_dict["content"] = response["result"]
        return message_dict

    def _build_system_prompt_from_messages(self, messages: List[BaseMessage]) -> Optional[str]:
        system_message_content_list: List[str] = []
        for msg in messages:
            if isinstance(msg, SystemMessage):
                if isinstance(msg.content, str):
                    system_message_content_list.append(msg.content)
                else:
                    raise TypeError
        if len(system_message_content_list) > 0:
            return "\n".join(system_message_content_list)
        else:
            return None

    def _convert_messages_to_dicts(self, messages: List[BaseMessage]) -> List[dict]:
        erniebot_messages = []
        for msg in messages:
            if isinstance(msg, SystemMessage):
                # Ignore system messages, as we handle them elsewhere.
                continue
            eb_msg = self._convert_message_to_dict(msg)
            erniebot_messages.append(eb_msg)
        return erniebot_messages

    @staticmethod
    def _convert_dict_to_message(message_dict: _MessageDict) -> BaseMessage:
        role = message_dict["role"]
        if role == "user":
            return HumanMessage(content=message_dict["content"])
        elif role == "assistant":
            content = message_dict["content"] or ""
            if message_dict.get("function_call"):
                additional_kwargs = {"function_call": dict(message_dict["function_call"])}
            else:
                additional_kwargs = {}
            return AIMessage(content=content, additional_kwargs=additional_kwargs)
        elif role == "function":
            return FunctionMessage(content=message_dict["content"], name=message_dict["name"])
        else:
            return ChatMessage(content=message_dict["content"], role=role)

    @staticmethod
    def _convert_message_to_dict(message: BaseMessage) -> _MessageDict:
        message_dict: _MessageDict
        if isinstance(message, ChatMessage):
            message_dict = {"role": message.role, "content": message.content}
        elif isinstance(message, HumanMessage):
            message_dict = {"role": "user", "content": message.content}
        elif isinstance(message, AIMessage):
            message_dict = {"role": "assistant", "content": message.content}
            if "function_call" in message.additional_kwargs:
                message_dict["function_call"] = message.additional_kwargs["function_call"]
                if message_dict["content"] == "":
                    message_dict["content"] = None
        elif isinstance(message, FunctionMessage):
            message_dict = {
                "role": "function",
                "content": message.content,
                "name": message.name,
            }
        else:
            raise TypeError(f"Got unknown type {message}")

        return message_dict



from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Type

from pydantic import Field

from erniebot_agent.memory.messages import AIMessage, HumanMessage, Message
from erniebot_agent.tools.schema import ToolParameterView

from .base import Tool


class CurrentTimeToolOutputView(ToolParameterView):
    current_time: str = Field(description="当前时间")


class CurrentTimeTool(Tool):
    description: str = "CurrentTimeTool 用于获取当前时间"
    ouptut_type: Type[ToolParameterView] = CurrentTimeToolOutputView

    async def __call__(self) -> Dict[str, str]:
        return {"current_time": datetime.strftime(datetime.now(), "%Y年%m月%d日 %H时%M分%S秒")}

    @property
    def examples(self) -> List[Message]:
        return [
            HumanMessage("现在几点钟了"),
            AIMessage(
                "",
                function_call={
                    "name": self.tool_name,
                    "thoughts": f"用户想知道现在几点了，我可以使用{self.tool_name}来获取当前时间，并从其中获得当前小时时间。",
                    "arguments": "{}",
                },
                token_usage={"prompt_tokens": 5, "completion_tokens": 7},  # For test only
            ),
            HumanMessage("现在是什么时候？"),
            AIMessage(
                "",
                function_call={
                    "name": self.tool_name,
                    "thoughts": f"用户想知道现在几点了，我可以使用{self.tool_name}来获取当前时间",
                    "arguments": "{}",
                },
                token_usage={"prompt_tokens": 5, "completion_tokens": 7},  # For test only
            ),
        ]



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, TypeVar, Union

from typing_extensions import Literal

from erniebot_agent.file import File, protocol
from erniebot_agent.memory import AIMessage, Message
from erniebot_agent.memory.messages import PluginInfo


@dataclass
class ToolAction(object):
    """A tool calling action for an agent to execute."""

    tool_name: str
    tool_args: str


@dataclass
class PluginAction(object):  # save for plugins that can be planned
    """A plugin calling action for an agent to execute."""

    plugin_name: str
    finish_reason: str


class ToolInfo(Dict):
    tool_name: str
    tool_args: str


@dataclass
class NullAction(object):
    """This class indicates that no action shall be taken by the agent."""


NULL_ACTION = NullAction()


# Note: save for plugins that can be planned
AgentAction = Union[ToolAction, PluginAction, NullAction]
PlanableAgentAction = Union[ToolAction, PluginAction]


@dataclass
class AgentPlan(object):
    """A plan that contains a list of actions."""

    actions: List[PlanableAgentAction]


@dataclass
class LLMResponse(object):
    """A response from an LLM."""

    message: AIMessage


@dataclass
class ToolResponse(object):
    """A response from a tool."""

    json: str
    input_files: List[File]
    output_files: List[File]


_IT = TypeVar("_IT", bound=Dict)
_RT = TypeVar("_RT")


@dataclass
class AgentStep(Generic[_IT, _RT]):
    """A step taken by an agent."""

    info: _IT
    result: _RT


@dataclass
class AgentStepWithFiles(AgentStep[_IT, _RT]):
    """A step taken by an agent involving file input and output."""

    input_files: List[File]
    output_files: List[File]

    @property
    def files(self) -> List[File]:
        return [*self.input_files, *self.output_files]


@dataclass
class ToolStep(AgentStepWithFiles[ToolInfo, Any]):
    """A step taken by an agent that calls a tool."""


@dataclass
class PluginStep(AgentStepWithFiles[PluginInfo, str]):
    """A step taken by an agent that calls a plugin."""


class _NullInfo(Dict):
    pass


class _NullResult(object):
    pass


@dataclass
class NoActionStep(AgentStep[_NullInfo, _NullResult]):
    """A step taken by an agent that performs no action and gives no result."""


NO_ACTION_STEP = NoActionStep(info=_NullInfo(), result=_NullResult())


class EndInfo(Dict):
    end_reason: str
    extra_info: str  # json format


@dataclass
class EndStep(AgentStep[EndInfo, None]):
    """A step taken by an agent that ends whole run."""


DEFAULT_FINISH_STEP = EndStep(info=EndInfo(end_reason="FINISHED"), result=None)


@dataclass
class AgentResponse(object):
    """The final response from an agent."""

    text: str
    chat_history: List[Message]
    steps: List[AgentStep]
    status: Union[Literal["FINISHED"], Literal["STOPPED"]]

    @functools.cached_property  # lazy and prevent extra fime from multiple calls
    def annotations(self) -> Dict[str, List]:
        annotations = self._get_annotations()

        return annotations

    def _get_annotations(self) -> Dict[str, List]:
        # 1. split the text into parts and add file id to each part
        file_ids = protocol.extract_file_ids(self.text)

        places = []
        for file_id in file_ids:
            # remote file-id & local file-id may have different length.
            # TODO(shiyutang): in case of multiple same file_id
            places.append((self.text.index(file_id), len(file_id)))
        else:
            sorted(places, key=lambda x: x[0])

        split_text_list = []
        prev_idx = 0
        for place in places:
            file_start_index, file_len = place
            split_text_list.append(self.text[prev_idx:file_start_index])
            split_text_list.append(self.text[file_start_index : file_start_index + file_len])
            prev_idx = file_start_index + file_len
        else:
            split_text_list.append(self.text[prev_idx:])

        # 2. parse text to dict
        annotations: Dict = {"content_parts": []}

        for file_id in split_text_list:
            if file_id in file_ids:
                for step in self.steps:
                    if isinstance(step, AgentStepWithFiles):
                        for file in step.files:
                            if file_id == file.id:
                                file_meta = file.to_dict()
                                annotations["content_parts"].append(file_meta)
            else:
                annotations["content_parts"].append({"text": file_id})

        return annotations



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The file module provides a series of classes for managing files,
which facilitate user interaction with the agent,
including the File base class and its subclasses, FileManager, GlobalFileManagerHandler,
and RemoteFileClient, which interacts with remote file servers.

It is recommended to use the GlobalFileManagerHandler to initialize the FileManager
and obtain the global FileManager at the beginning of the event loop.
Afterwards, you can simply use this global FileManager to perform operations such as adding,
deleting, and querying files, as well as obtaining files generated by the agent.


A few notes about this submodule:

- If you do not set environment variable `AISTUDIO_ACCESS_TOKEN`, it will be under default setting.

- Method `GlobalFileManagerHandler().configure()` can only be called **once** at the beginning.

- When you want to get a file manger, you can use method `GlobalFileManagerHandler().get()`.

- The lifecycle of the `FileManager` class is synchronized with the event loop.

- `FileManager` class is Noncopyable.

- If you want to get the content of `File` object, you can use `read_contents`
  and use `write_contents_to` create the file to location you want.

- We do **not** recommend you to create `File` object yourself.

Examples:
    >>> from erniebot_agent.file import GlobalFileManagerHandler
    >>> async def demo_function():
    >>>     # need to use at the beginning of event loop
    >>>     GlobalFileManagerHandler().configure(save_dir='your_path')
    >>>     file_manager = GlobalFileManagerHandler().get()
    >>>     local_file = await file_manager.create_file_from_path(file_path='your_path', file_type='local')

    >>>     file = file_manager.look_up_file_by_id(file_id='your_file_id')
    >>>     file_content = await file.read_contents() # get file content(bytes)
    >>>     await local_file.write_contents_to('your_willing_path') # save to location you want
"""

from erniebot_agent.file.file_manager import File, FileManager, get_default_file_manager
from erniebot_agent.file.global_file_manager_handler import GlobalFileManagerHandler
from erniebot_agent.file.remote_file import AIStudioFileClient



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Generic, List, Optional, TypeVar, final

from erniebot_agent.file.base import BaseFile

_T = TypeVar("_T", bound=BaseFile)


@final
class FileRegistry(Generic[_T]):
    """
    Singleton class for managing file registration.


    Methods:
        register_file: Register a file in the registry.
        unregister_file: Unregister a file from the registry.
        look_up_file: Look up a file by its ID in the registry.
        list_files: Get a list of all registered files.

    """

    def __init__(self) -> None:
        super().__init__()
        self._id_to_file: Dict[str, _T] = {}

    def register_file(self, file: _T, *, allow_overwrite: bool = False) -> None:
        """
        Register a file in the registry.

        Args:
            file (File): The file object to register.
            allow_overwrite (bool): Allow overwriting if a file with the same ID is already registered.

        Returns:
            None

        Raises:
            RuntimeError: If the file ID is already registered and allow_overwrite is False.

        """
        file_id = file.id
        if file_id in self._id_to_file:
            if not allow_overwrite:
                raise ValueError(f"File with ID {repr(file_id)} is already registered.")
        self._id_to_file[file_id] = file

    def unregister_file(self, file: _T) -> None:
        """
        Unregister a file from the registry.

        Args:
            file (File): The file object to unregister.

        Returns:
            None

        Raises:
            RuntimeError: If the file ID is not registered.

        """
        file_id = file.id
        if file_id not in self._id_to_file:
            raise ValueError(f"File with ID {repr(file_id)} is not registered.")
        self._id_to_file.pop(file_id)

    def look_up_file(self, file_id: str) -> Optional[_T]:
        """
        Look up a file by its ID in the registry.

        Args:
            file_id (str): The ID of the file to look up.

        Returns:
            Optional[File]: The File object if found, or None if not found.

        """
        return self._id_to_file.get(file_id, None)

    def list_files(self) -> List[_T]:
        """
        Get a list of all registered files.

        Returns:
            List[File]: The list of registered File objects.

        """
        return list(self._id_to_file.values())



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional, Tuple, Union


def get_global_access_token() -> Optional[str]:
    return _get_val_from_env_var("EB_AGENT_ACCESS_TOKEN")


def get_global_save_dir() -> Optional[str]:
    return _get_val_from_env_var("EB_AGENT_SAVE_DIR")


def get_logging_level() -> Optional[str]:
    return _get_val_from_env_var("EB_AGENT_LOGGING_LEVEL")


def get_logging_file_path() -> Optional[str]:
    return _get_val_from_env_var("EB_AGENT_LOGGING_FILE")


def get_global_aksk() -> Tuple[Union[str, None], Union[str, None]]:
    return (_get_val_from_env_var("EB_AGENT_AK"), _get_val_from_env_var("EB_AGENT_SK"))


def _get_val_from_env_var(env_var: str) -> Optional[str]:
    return os.getenv(env_var)



# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from erniebot_agent.agents.callback.handlers.base import CallbackHandler
from erniebot_agent.agents.callback.handlers.logging_handler import LoggingHandler


def get_default_callbacks() -> List[CallbackHandler]:
    """Get default callback handlers."""
    return [LoggingHandler()]


def get_no_ellipsis_callback() -> List[CallbackHandler]:
    return [LoggingHandler()]



from __future__ import annotations

from typing import Dict, List, Type

from pydantic import Field

from erniebot_agent.memory.messages import AIMessage, HumanMessage, Message
from erniebot_agent.tools.schema import ToolParameterView

from .base import Tool


class CalculatorToolInputView(ToolParameterView):
    math_formula: str = Field(description='标准的数学公式，例如："2+3"、"3 - 4 * 6", "(3 + 4) * (6 + 4)" 等。 ')


class CalculatorToolOutputView(ToolParameterView):
    formula_result: float = Field(description="数学公式计算的结果")


class CalculatorTool(Tool):
    description: str = "CalculatorTool用于执行数学公式计算"
    input_type: Type[ToolParameterView] = CalculatorToolInputView
    ouptut_type: Type[ToolParameterView] = CalculatorToolOutputView

    async def __call__(self, math_formula: str) -> Dict[str, float]:
        return {"formula_result": eval(math_formula)}

    @property
    def examples(self) -> List[Message]:
        return [
            HumanMessage("请告诉我三加六等于多少？"),
            AIMessage(
                "",
                function_call={
                    "name": self.tool_name,
                    "thoughts": f"用户想知道3加6等于多少，我可以使用{self.tool_name}工具来计算公式，其中`math_formula`字段的内容为：'3+6'。",
                    "arguments": '{"math_formula": "3+6"}',
                },
                token_usage={
                    "prompt_tokens": 5,
                    "completion_tokens": 7,
                },  # TODO: Functional AIMessage will not add in the memory, will it add token_usage?
            ),
            HumanMessage("一加八再乘以5是多少？"),
            AIMessage(
                "",
                function_call={
                    "name": self.tool_name,
                    "thoughts": f"用户想知道1加8再乘5等于多少，我可以使用{self.tool_name}工具来计算公式，"
                    "其中`math_formula`字段的内容为：'(1+8)*5'。",
                    "arguments": '{"math_formula": "(1+8)*5"}',
                },
                token_usage={"prompt_tokens": 5, "completion_tokens": 7},  # For test only
            ),
            HumanMessage("我想知道十二除以四再加五等于多少？"),
            AIMessage(
                "",
                function_call={
                    "name": self.tool_name,
                    "thoughts": f"用户想知道12除以4再加5等于多少，我可以使用{self.tool_name}工具来计算公式，"
                    "其中`math_formula`字段的内容为：'12/4+5'。",
                    "arguments": '{"math_formula": "12/4+5"}',
                },
                token_usage={"prompt_tokens": 5, "completion_tokens": 7},  # For test only
            ),
        ]



import base64
import inspect
import logging
import typing
from copy import deepcopy
from typing import Any, Dict, Optional, Type, no_type_check

from openapi_spec_validator import validate
from openapi_spec_validator.readers import read_from_filename
from requests import Response

from erniebot_agent.file import File, FileManager
from erniebot_agent.file.protocol import (
    FilePurpose,
    is_local_file_id,
    is_remote_file_id,
)
from erniebot_agent.tools.schema import (
    ToolParameterView,
    get_args,
    get_typing_list_type,
)
from erniebot_agent.utils.common import get_file_suffix, import_module, is_json_response
from erniebot_agent.utils.exceptions import RemoteToolError

_logger = logging.getLogger(__name__)


def tool_response_contains_file(element: Any):
    if isinstance(element, str):
        if is_local_file_id(element) or is_remote_file_id(element):
            return True
    elif isinstance(element, dict):
        for val in element.values():
            if tool_response_contains_file(val):
                return True
    elif isinstance(element, list):
        for val in element:
            if tool_response_contains_file(val):
                return True


def validate_openapi_yaml(yaml_file: str) -> bool:
    """do validation on the yaml file

    Args:
        yaml_file (str): the path of yaml file

    Returns:
        bool: whether yaml file is valid
    """
    yaml_dict = read_from_filename(yaml_file)[0]
    try:
        validate(yaml_dict)
        return True
    except Exception as e:  # type: ignore
        _logger.error(e)
        _logger.error(
            "You can edit your openapi.yaml file in https://editor.swagger.io/ "
            "which is more friendly for you to find the issue."
        )
        return False


def get_file_info_from_param_view(
    param_view: Optional[Type[ToolParameterView]] = None,
) -> Dict[str, Dict[str, str]]:
    """get file names from tool parameter view

    Args:
        param_view (ToolParameterView): the ToolParameterView pydantic class

    Returns:
        List[str]: the names of file
    """
    if param_view is None:
        return {}

    file_infos: Dict[str, Any] = {}
    for key in param_view.model_fields.keys():
        model_field = param_view.model_fields[key]

        list_base_annotation = get_typing_list_type(model_field.annotation)
        json_schema_extra: dict = typing.cast(dict, model_field.json_schema_extra)

        if list_base_annotation is not None:
            if list_base_annotation == "object":
                # get base type
                arg_type = get_args(model_field.annotation)[0]
                sub_file_infos = get_file_info_from_param_view(arg_type)
                if len(sub_file_infos) > 0:
                    file_infos[key] = sub_file_infos
                continue

            if "array_items_schema" in json_schema_extra:
                json_schema_extra = json_schema_extra["array_items_schema"]
                if not isinstance(json_schema_extra, dict):
                    raise RemoteToolError(
                        f"<array_items_schema> field must be dict type in model_field<{key}> "
                        f"with the field_info<{model_field}>. Please check the format of yaml "
                        "in current tool.",
                        stage="Output parsing",
                    )

        elif model_field.annotation is not None and issubclass(model_field.annotation, ToolParameterView):
            sub_file_infos = get_file_info_from_param_view(model_field.annotation)
            if len(sub_file_infos) > 0:
                file_infos[key] = sub_file_infos
            continue

        if json_schema_extra and json_schema_extra.get("format", None) in [  # type: ignore
            "byte",
            "binary",
        ]:
            file_infos[key] = deepcopy(json_schema_extra)
    return file_infos


async def parse_file_from_json_response(
    json_data: dict, file_manager: FileManager, param_view: Type[ToolParameterView], tool_name: str
):
    if param_view is None:
        return {}

    file_infos: Dict[str, Any] = {}
    for key in param_view.model_fields.keys():
        model_field = param_view.model_fields[key]

        # to avoid: yaml schema is not matched with json response schema
        if key not in json_data:
            continue

        list_base_annotation = get_typing_list_type(model_field.annotation)
        if list_base_annotation == "object":
            # get base type
            arg_type = get_args(model_field.annotation)[0]
            sub_file_infos_list = []
            for json_item in json_data[key]:
                sub_file_infos = await parse_file_from_json_response(
                    json_item, file_manager=file_manager, param_view=arg_type, tool_name=tool_name
                )
                if sub_file_infos:
                    sub_file_infos_list.append(sub_file_infos)
            if sub_file_infos_list:
                file_infos[key] = sub_file_infos_list
        elif inspect.isclass(model_field.annotation) and issubclass(
            model_field.annotation, ToolParameterView
        ):
            sub_file_infos = await parse_file_from_json_response(
                json_data[key],
                file_manager=file_manager,
                param_view=model_field.annotation,
                tool_name=tool_name,
            )
            if sub_file_infos:
                file_infos[key] = sub_file_infos
        else:
            json_schema_extra = model_field.json_schema_extra
            if not isinstance(json_schema_extra, dict):
                raise RemoteToolError(
                    f"<json_schema_extra> field must be dict type in model_field<{key}> "
                    f"with the field_info<{model_field}>. Please check the format of yaml in current tool.",
                    stage="Output parsing",
                )

            format = json_schema_extra.get("format", None)
            if format in ["byte", "binary"]:
                content = json_data[key]
                if format == "byte":
                    content = base64.b64decode(content)

                mime_type = json_schema_extra.get("x-ebagent-file-mime-type", None)
                if mime_type is not None and not isinstance(mime_type, str):
                    raise RemoteToolError(
                        f"x-ebagent-file-mime-type value must be None or string in key<{key}>, "
                        f"but receive ({type(mime_type)})<{mime_type}>",
                        stage="Output parsing",
                    )

                suffix = get_file_suffix(mime_type)
                file = await file_manager.create_file_from_bytes(
                    content,
                    filename=f"test{suffix}",
                    file_purpose="assistants_output",
                    file_metadata={"tool_name": tool_name},
                )
                file_infos[key] = file.id
    return file_infos


async def create_file_from_data(
    content,
    format: str,
    mime_type: str,
    file_manager: FileManager,
    file_purpose: FilePurpose = "assistants_output",
    file_metadata={},
) -> File:
    file_suffix = get_file_suffix(mime_type)
    if format == "byte":
        content = base64.b64decode(content)

    return await file_manager.create_file_from_bytes(
        content,
        f"tool-{file_suffix}",
        file_purpose=file_purpose,
        file_metadata=file_metadata,
    )


async def get_content_by_file_id(file_id: str, format: str, file_manager: FileManager) -> str:
    file_id = file_id.replace("<file>", "").replace("</file>", "")
    file = file_manager.look_up_file_by_id(file_id)
    byte_str = await file.read_contents()
    if format == "byte":
        byte_str = base64.b64encode(byte_str)

    return byte_str.decode()


def is_file_config(json_schema_extra: dict) -> bool:
    """check wheter is file-config

    Args:
        json_schema_extra (dict): the config from yaml file

    Returns:
        bool: whether is file-config
    """
    return json_schema_extra.get("format", None) in ["byte", "binary"]


@no_type_check
async def parse_json_request(
    view: Type[ToolParameterView], json_dict, file_manager: FileManager
) -> Dict[str, Any]:
    result = {}
    for field_name, model_field in view.model_fields.items():
        list_type = get_typing_list_type(model_field.annotation)

        if list_type is None and model_field.json_schema_extra is not None:
            if not isinstance(model_field.json_schema_extra, dict):
                raise RemoteToolError(
                    f"`json_schema_extra` field must be dict type, but got "
                    f"<{type(model_field.json_schema_extra)}> in model_field<{field_name}>. "
                    "Please check the format of yaml in current tool."
                )

            if model_field.annotation == str and is_file_config(model_field.json_schema_extra):
                format = model_field.json_schema_extra.get("format", None)

                if format is not None:
                    file_content = await get_content_by_file_id(
                        json_dict[field_name], format=format, file_manager=file_manager
                    )
                    result[field_name] = file_content
            elif issubclass(model_field.annotation, ToolParameterView):
                files = await parse_json_request(
                    model_field.annotation,
                    json_dict[field_name],
                    file_manager=file_manager,
                )
                if len(files) > 0:
                    result[field_name] = files

        else:
            array_json_schema = model_field.json_schema_extra.get("array_items_schema", None)
            sub_class = get_args(model_field.annotation)[0]
            if list_type == "string" and is_file_config(array_json_schema):
                format = array_json_schema["format"]
                files = []
                for file_id in json_dict[field_name]:
                    file_content = await get_content_by_file_id(
                        file_id,
                        format=format,
                        file_manager=file_manager,
                    )
                    files.append(file_content)

                result[field_name] = files

            elif list_type == "object":
                sub_file_result = []
                for file_dict in json_dict[field_name]:
                    sub_file = await parse_json_request(sub_class, file_dict, file_manager)
                    if len(sub_file) > 0:
                        sub_file_result.append(sub_file)

                if len(sub_file_result) > 0:
                    result[field_name] = sub_file_result

        if field_name in result:
            json_dict.pop(field_name, None)
        elif field_name not in result and field_name in json_dict:
            result[field_name] = json_dict.pop(field_name)

        fixed_value = model_field.json_schema_extra.get("x-ebagent-fixed-value", None)
        if fixed_value:
            result[field_name] = fixed_value
    result.update(json_dict)
    return result


@no_type_check
async def parse_json_response(
    view: Type[ToolParameterView], json_dict, file_manager: FileManager, file_metadata: Dict[str, str]
) -> Dict[str, Any]:
    result = {}
    for field_name, model_field in view.model_fields.items():
        list_type = get_typing_list_type(model_field.annotation)

        if list_type is None and model_field.json_schema_extra is not None:
            if not isinstance(model_field.json_schema_extra, dict):
                raise RemoteToolError(
                    f"`json_schema_extra` field must be dict type, but got "
                    f"<{type(model_field.json_schema_extra)}> in model_field<{field_name}>. "
                    "Please check the format of yaml in current tool."
                )

            if model_field.annotation == str and "x-ebagent-file-mime-type" in model_field.json_schema_extra:
                format = model_field.json_schema_extra.get("format", None)
                mime_type = model_field.json_schema_extra.get("x-ebagent-file-mime-type", None)

                if format is not None and mime_type is not None:
                    file = await create_file_from_data(
                        json_dict[field_name],
                        format=format,
                        mime_type=mime_type,
                        file_manager=file_manager,
                        file_metadata=file_metadata,
                    )
                    result[field_name] = file.id
            elif issubclass(model_field.annotation, ToolParameterView):
                files = await parse_json_response(
                    model_field.annotation,
                    json_dict[field_name],
                    file_manager=file_manager,
                    file_metadata=file_metadata,
                )
                if len(files) > 0:
                    result[field_name] = files

        else:
            array_json_schema = model_field.json_schema_extra.get("array_items_schema", None)
            sub_class = get_args(model_field.annotation)[0]
            if (
                list_type == "string"
                and array_json_schema is not None
                and array_json_schema.get("x-ebagent-file-mime-type", None)
            ):
                format = array_json_schema["format"]
                mime_type = array_json_schema["x-ebagent-file-mime-type"]
                files = []
                for file_content in json_dict[field_name]:
                    file = await create_file_from_data(
                        file_content,
                        format=format,
                        mime_type=mime_type,
                        file_manager=file_manager,
                        file_metadata=file_metadata,
                    )
                    files.append(file.id)
                result[field_name] = files
            elif list_type == "object":
                sub_file_result = []
                for file_dict in json_dict[field_name]:
                    sub_file = await parse_json_response(sub_class, file_dict, file_manager, file_metadata)
                    if len(sub_file) > 0:
                        sub_file_result.append(sub_file)

                if len(sub_file_result) > 0:
                    result[field_name] = sub_file_result

        if field_name in result:
            json_dict.pop(field_name, None)
        elif field_name not in result and field_name in json_dict:
            result[field_name] = json_dict.pop(field_name)

        fixed_value = model_field.json_schema_extra.get("x-ebagent-fixed-value", None)
        if fixed_value:
            result[field_name] = fixed_value

    result.update(json_dict)
    return result


async def parse_response(
    response: Response,
    file_manager: FileManager,
    file_metadata: Dict[str, str] = {},
    tool_parameter_view: Optional[Type[ToolParameterView]] = None,
) -> Dict[str, Any]:
    # 1. parse file by `Content-Disposition`
    content_disposition = response.headers.get("Content-Disposition", None)
    if content_disposition is not None:
        file_name = response.headers["Content-Disposition"].split("filename=")[1]
        local_file = await file_manager.create_file_from_bytes(
            response.content, file_name, file_purpose="assistants_output", file_metadata=file_metadata
        )
        return {"file": local_file.id}

    if is_json_response(response):
        if tool_parameter_view is None:
            return response.json()

        return await parse_json_response(
            tool_parameter_view, response.json(), file_manager=file_manager, file_metadata=file_metadata
        )

    # 3. parse file by content_type
    content_type = response.headers.get("Content-Type", None)
    if content_type is not None:
        file_suffix = get_file_suffix(content_type)
        local_file = await file_manager.create_file_from_bytes(
            response.content,
            f"tool-{file_suffix}",
            file_purpose="assistants_output",
            file_metadata=file_metadata,
        )
        return {"file": local_file.id}

    raise RemoteToolError(
        "Can not parse file from response: the type of data from response is not json, "
        "and can not find `Content-Disposition` or `Content-Type` field from response header.",
        stage="Output parsing",
    )


def get_fastapi_openapi(app):
    """get openapi dict of fastapi application

    refer to: https://github.com/tiangolo/fastapi/issues/3424#issuecomment-1283484665

    Args:
        app (FastAPI): the instance of fastapi application
    """
    fastapi = import_module(
        "fastapi",
        "Could not import fastapi or uvicorn python package. Please install it "
        "with `pip install fastapi`.",
    )

    if not app.openapi_schema:
        app.openapi_schema = fastapi.openapi.utils.get_openapi(
            title=app.title,
            version=app.version,
            openapi_version=app.openapi_version,
            description=app.description,
            terms_of_service=app.terms_of_service,
            contact=app.contact,
            license_info=app.license_info,
            routes=app.routes,
            tags=app.openapi_tags,
            servers=app.servers,
        )

        # remove validation response
        for _, method_item in app.openapi_schema.get("paths", {}).items():
            for _, param in method_item.items():
                responses = param.get("responses")
                # remove 422 response, also can remove other status code
                if "422" in responses:
                    del responses["422"]

        # remove Validation Schema
        schemas = deepcopy(app.openapi_schema["components"]["schemas"])
        for key in list(schemas.keys()):
            if "ValidationError" in key:
                schemas.pop(key)

        app.openapi_schema["components"]["schemas"] = schemas
